from urllib import parse

from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

import xmatters.errors as err
import xmatters.utils as util


class Connection(object):
    def __init__(self, base_url, session, **kwargs):
        self.base_url = base_url
        self.session = session
        p_url = parse.urlparse(self.base_url)
        self.api_path = p_url.path
        self.instance_url = 'https://{}'.format(p_url.netloc)
        self.timeout = kwargs.get('timeout')
        if kwargs.get('max_retries'):
            self.max_retries = kwargs.get('max_retries')

    def get(self, url, params=None):
        return self.request('GET', url=url, params=params)

    def post(self, url, data):
        return self.request('POST', url=url, data=data)

    def delete(self, url):
        return self.request('DELETE', url=url)

    def request(self, method, url, data=None, params=None):
        if params:
            # use api max limit if one not specified
            if 'limit' in params.keys() and params.get('limit') is None:
                params['limit'] = util.MAX_API_LIMIT
            r = self.session.request(method=method, url=url, params={k: v for k, v in params.items() if v},
                                     timeout=self.timeout)
        elif data:
            r = self.session.request(method=method, url=url, json=data, timeout=self.timeout)
        else:
            r = self.session.request(method=method, url=url, timeout=self.timeout)
        data = r.json()
        if not r.ok or r.status_code == 204:
            raise err.ErrorFactory.compose(None, data, err.ApiError)
        return data

    @property
    def max_retries(self):
        return self._max_retries

    @max_retries.setter
    def max_retries(self, retries):
        self._max_retries = retries
        retry = Retry(total=retries,
                      backoff_factor=0.1,
                      status_forcelist=[500, 502, 503, 504])
        retry_adapter = HTTPAdapter(max_retries=retry)
        self.session.mount('https://', retry_adapter)

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class ApiBridge(object):
    """ Base for api objects that need to make requests """

    def __init__(self, parent, data=None):
        if hasattr(parent, 'con') and not parent.con:
            raise err.AuthorizationError('authentication not provided')

        self.con = parent.con
        self.self_url = None
        self_link = data.get('links', {}).get('self') if data else None
        self.self_url = '{}{}'.format(self.con.instance_url, self_link) if self_link else None

    def build_url(self, endpoint):
        # don't do anything if endpoint is full path
        if self.con.instance_url in endpoint:
            return endpoint

        if endpoint.startswith(self.con.api_path):
            url_prefix = self.con.instance_url
        elif self.self_url:
            url_prefix = self.self_url
        else:
            url_prefix = self.con.base_url
        return '{}{}'.format(url_prefix, endpoint)

    @staticmethod
    def process_time_param(param):
        """
        Formats iso-formatted date (or date and time) from provided timezone to UTC timezone.
        If no timezone is specified, local timezone is used.
        :param param: iso-formatted date, or date and time
        :type param: str
        :return: date & time with utc offset applied
        :rtype: str or None
        """
        if param:
            return util.TimeAttribute(param).isoformat_utc()



import xmatters.connection
import xmatters.xm_objects.common


class Site(xmatters.connection.ApiBridge):
    def __init__(self, parent, data):
        super(Site, self).__init__(parent, data)
        self.id = data.get('id')
        self.address1 = data.get('address1')
        self.address2 = data.get('address2')
        self.city = data.get('city')
        self.country = data.get('country')
        self.external_key = data.get('externalKey')
        self.externally_owned = data.get('externallyOwned')
        self.language = data.get('language')
        self.latitude = data.get('latitude')
        links = data.get('links')
        self.links = xmatters.xm_objects.common.SelfLink(self, links) if links else None
        self.longitude = data.get('longitude')
        self.name = data.get('name')
        self.postal_code = data.get('postalCode')
        self.state = data.get('state')
        self.status = data.get('status')
        self.timezone = data.get('timezone')

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()



import xmatters.factories
import xmatters.utils as utils
from xmatters.xm_objects.common import Recipient, SelfLink, Pagination
from xmatters.xm_objects.roles import Role
from xmatters.connection import ApiBridge
import xmatters.xm_objects.groups


class Person(Recipient):
    _endpoints = {'get_devices': '/devices',
                  'roles': '?embed=roles',
                  'get_supervisors': '/supervisors',
                  'supervisors': '?embed=supervisors',
                  'get_groups': '/group-memberships'}

    def __init__(self, parent, data):
        super(Person, self).__init__(parent, data)
        self.first_name = data.get('firstName')
        self.last_name = data.get('lastName')
        self.license_type = data.get('licenseType')
        self.language = data.get('language')
        self.timezone = data.get('timezone')
        self.web_login = data.get('webLogin')
        self.phone_login = data.get('phoneLogin')
        self.phone_pin = data.get('phonePin')
        self.properties = data.get('properties', {})
        last_login = data.get('lastLogin')
        self.last_login = utils.TimeAttribute(last_login) if last_login else None
        when_created = data.get('whenCreated')
        self.when_created = utils.TimeAttribute(when_created) if when_created else None
        when_updated = data.get('whenUpdated')
        self.when_updated = utils.TimeAttribute(when_updated) if when_updated else None
        links = data.get('links')
        self.links = SelfLink(self, data) if links else None

    @property
    def roles(self):
        url = self.build_url(self._endpoints.get('roles'))
        data = self.con.get(url)
        roles = data.get('roles', {})
        return Pagination(self, roles, Role) if roles.get('data') else []

    @property
    def devices(self):
        return self.get_devices()

    @property
    def supervisors(self):
        return self.get_supervisors()

    def get_supervisors(self, offset=None, limit=None):
        params = {'offset': offset,
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_supervisors'))
        s = self.con.get(url, params=params)
        return Pagination(self, s, Person) if s.get('data') else []

    def get_devices(self, offset=None, limit=None, phone_number_format=None, at=None):
        params = {'phoneNumberFormat': phone_number_format,
                  'at': self.process_time_param(at),
                  'offset': offset,
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_devices'))
        devices = self.con.get(url, params=params)
        return Pagination(self, devices, xmatters.factories.DeviceFactory) if devices.get('data') else []

    def get_groups(self, offset=None, limit=None):
        params = {'offset': offset,
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_groups'))
        groups = self.con.get(url, params=params)
        return Pagination(self, groups, xmatters.xm_objects.groups.GroupMembership) if groups.get('data') else []

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.target_name)

    def __str__(self):
        return self.__repr__()


class PersonReference(ApiBridge):
    def __init__(self, parent, data):
        super(PersonReference, self).__init__(parent, data)
        self.id = data.get('id')
        self.target_name = data.get('targetName')
        self.first_name = data.get('firstName')
        self.last_name = data.get('lastName')
        self.recipient_type = data.get('recipientType')
        links = data.get('links')
        self.links = SelfLink(self, links) if links else None

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.target_name)

    def __str__(self):
        return self.__repr__()



import xmatters.connection
import xmatters.factories as factory
import xmatters.xm_objects.events as events
import xmatters.xm_objects.plans as plans
import xmatters.xm_objects.scenarios
from xmatters.xm_objects.device_names import TargetDeviceNameSelector
from xmatters.xm_objects.common import Pagination, Recipient, PropertyDefinition, SelfLink


class FormReference(object):
    def __init__(self, data):
        self.id = data.get('id')
        self.name = data.get('name')

    def __repr__(self):
        return '<{}>'.format(self.__class__)

    def __str__(self):
        return self.__repr__()


class SectionValue(object):

    def __init__(self, data):
        self.id = data.get('id')
        self.value = data.get('value')
        self.visible = data.get('visible')


class SenderOverrides(object):
    def __init__(self, data):
        caller_id = data.get('callerId')
        self.caller_id = SectionValue(caller_id) if caller_id else None
        display_name = data.get('displayName')
        self.display_name = SectionValue(display_name) if display_name else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class FormSection(xmatters.connection.ApiBridge):
    def __init__(self, parent, data):
        super(FormSection, self).__init__(parent, data)
        self.id = data.get('id')
        form = data.get('form')
        self.form = FormReference(form) if form else None
        self.title = data.get('title')
        self.type = data.get('type')
        self.visible = data.get('visible')
        self.collapsed = data.get('collapsed')
        self.order_num = data.get('orderNum')

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.type)

    def __str__(self):
        return self.__repr__()


class IncidentSectionItem(object):
    def __init__(self, data):
        self.value = data.get('value')
        self.order_num = data.get('orderNum')
        self.visible = data.get('visible')
        self.required = data.get('required')

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class IncidentSection(FormSection):
    def __init__(self, parent, data):
        super(IncidentSection, self).__init__(parent, data)
        summary = data.get('summary')
        self.summary = IncidentSectionItem(summary) if summary else None
        description = data.get('description')
        self.description = IncidentSectionItem(description) if description else None
        severity = data.get('severity')
        self.severity = IncidentSectionItem(severity) if severity else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class ConferenceBridgeSection(FormSection):
    def __init__(self, parent, data):
        super(ConferenceBridgeSection, self).__init__(parent, data)
        self.bridge_type = data.get('bridgeType')

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.bridge_type)

    def __str__(self):
        return self.__repr__()


class CustomSectionItems(FormSection):

    def __init__(self, parent, data):
        super(CustomSectionItems, self).__init__(parent, data)
        items = data.get('items')
        self.items = [CustomSection(i) for i in items] if items else []

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class CustomSection(object):
    def __init__(self, data):
        self.id = data.get('id')
        form_section = data.get('formSection')
        self.form_section = FormReference(form_section) if form_section else None
        self.order_num = data.get('orderNum')
        self.required = data.get('required')
        self.multiline_text = data.get('multiLineText')
        self.visible = data.get('visible')
        self.include_in_callback = data.get('includeInCallback')
        property_definition = data.get('propertyDefinition')
        self.property_definition = PropertyDefinition(property_definition) if property_definition else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class DevicesSection(FormSection):
    def __init__(self, parent, data):
        super(DevicesSection, self).__init__(parent, data)
        tdns = data.get('targetDeviceNames', {})
        self.target_device_names = Pagination(self, tdns, TargetDeviceNameSelector) if tdns.get('data') else []

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class HandlingSection(FormSection):
    def __init__(self, parent, data):
        super(HandlingSection, self).__init__(parent, data)
        threshold = data.get('otherResponseCountThreshold')
        self.other_response_count_threshold = SectionValue(threshold) if threshold else None
        priority = data.get('priority')
        self.priority = SectionValue(priority) if priority else None
        expiration_in_minutes = data.get('expirationInMinutes')
        self.expiration_in_minutes = SectionValue(expiration_in_minutes) if expiration_in_minutes else None
        override = data.get('overrideDeviceRestrictions')
        self.override_device_restrictions = SectionValue(override) if override else None
        escalation_override = data.get('escalationOverride')
        self.escalation_override = SectionValue(escalation_override) if escalation_override else None
        bypass_phone_intro = data.get('bypassPhoneIntro')
        self.bypass_phone_intro = SectionValue(bypass_phone_intro) if bypass_phone_intro else None
        require_phone_password = data.get('requirePhonePassword')
        self.require_phone_password = SectionValue(require_phone_password) if require_phone_password else None
        voicemail_options = data.get('voicemailOptions')
        self.voicemail_options = events.VoicemailOptions(data) if voicemail_options else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class RecipientsSection(FormSection):
    def __init__(self, parent, data):
        super(RecipientsSection, self).__init__(parent, data)
        recipients = data.get('recipients', {})
        self.recipients = Pagination(self, recipients, Recipient) if recipients.get('data') else []

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class SenderOverridesSection(FormSection):
    def __init__(self, parent, data):
        super(SenderOverridesSection, self).__init__(parent, data)
        self.sender_overrides = SenderOverrides(data) if data else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class Form(xmatters.connection.ApiBridge):
    _endpoints = {'response_options': '?embed=responseOptions',
                  'get_response_options': '/response-options',
                  'get_sections': '{base_url}/forms/{form_id}/sections',
                  'recipients': '?embed=recipients',
                  'get_scenarios': '/scenarios'}

    def __init__(self, parent, data):
        super(Form, self).__init__(parent, data)
        self.id = data.get('id')
        self.form_id = data.get('formId')
        self.name = data.get('name')
        self.description = data.get('description')
        self.mobile_enabled = data.get('mobileEnabled')
        self.ui_enabled = data.get('uiEnabled')
        self.api_enabled = data.get('apiEnabled')
        sender_overrides = data.get('senderOverrides')
        self.sender_overrides = SenderOverrides(sender_overrides) if sender_overrides else None
        plan = data.get('plan')
        self.plan = plans.PlanReference(plan) if plan else None
        links = data.get('links')
        self.links = SelfLink(self, data) if links else None

    @property
    def response_options(self):
        return self.get_response_options()

    @property
    def recipients(self, params=None):
        url = self.build_url(self._endpoints.get('recipients'))
        recipients = self.con.get(url, params).get('recipients', {})
        return Pagination(self, recipients, factory.RecipientFactory) if recipients.get( 'data') else []

    def get_response_options(self, params=None):
        url = self.build_url(self._endpoints.get('get_response_options'))
        options = self.con.get(url, params)
        return Pagination(self, options, events.ResponseOption) if options.get(
            'data') else []

    def get_sections(self, offset=None, limit=None):
        params = {'offset': offset,
                  'limit': limit}
        url = self._endpoints.get('get_sections').format(base_url=self.con.base_url, form_id=self.id)
        s = self.con.get(url, params=params)
        return Pagination(self, s, factory.SectionFactory, limit=limit) if s.get(
            'data') else []

    # TODO: Test params
    def get_scenarios(self, search=None, operand=None, enabled_for=None, offset=None, limit=None):
        params = {'search': ' '.join(search) if search else None,
                  'operand': operand,
                  'enabledFor': enabled_for,
                  'offset': offset,
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_scenarios'))
        s = self.con.get(url, params=params)
        return Pagination(self, s, xmatters.xm_objects.scenarios.Scenario) if s.get(
            'data') else []

    # TODO: Test
    def create_scenario(self, data):
        url = self.build_url(self._endpoints.get('get_scenarios'))
        data = self.con.post(url, data=data)
        return xmatters.xm_objects.scenarios.Scenario(self, data) if data else None

    # TODO: Test
    def update_scenario(self, data):
        url = self.build_url(self._endpoints.get('get_scenarios'))
        data = self.con.post(url, data=data)
        return xmatters.xm_objects.scenarios.Scenario(self, data) if data else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()



import xmatters.factories as factory
import xmatters.xm_objects.forms
from xmatters.connection import ApiBridge
from xmatters.xm_objects.common import Pagination
from xmatters.xm_objects.conference_bridges import ConferenceBridge
from xmatters.xm_objects.device_types import DeviceTypes
from xmatters.xm_objects.dynamic_teams import DynamicTeam
from xmatters.xm_objects.event_supressions import EventSuppression
from xmatters.xm_objects.events import Event
from xmatters.xm_objects.groups import Group
from xmatters.xm_objects.import_jobs import Import
from xmatters.xm_objects.incidents import Incident
from xmatters.xm_objects.oncall import OnCall
from xmatters.xm_objects.oncall_summary import OnCallSummary
from xmatters.xm_objects.people import Person
from xmatters.xm_objects.plans import Plan
from xmatters.xm_objects.roles import Role
from xmatters.xm_objects.scenarios import Scenario
from xmatters.xm_objects.services import Service
from xmatters.xm_objects.sites import Site
from xmatters.xm_objects.subscription_forms import SubscriptionForm
from xmatters.xm_objects.subscriptions import Subscription
from xmatters.xm_objects.temporary_absences import TemporaryAbsence


class AuditsEndpoint(ApiBridge):
    """ Used to interact with '/audit' endpoint """

    def __init__(self, parent):
        """
        :param parent: XMSession instance
        :type parent: :class:`xmatters.session.XMSession`
        """
        super(AuditsEndpoint, self).__init__(parent)

        self._endpoints = {'get_audit': '/audits'}

    # TODO: update docstring
    def get_audit(self, event_id=None, audit_type=None, sort_order=None, at_time=None, from_time=None, to_time=None,
                  after_time=None, before_time=None, offset=None, limit=None):
        """
        Perform an audit on a specified event id.

        See `xMatters REST API Reference <https://help.xmatters.com/xmapi/>`_ for valid parameters.

        :param event_id: xMatters event id
        :type event_id: str
        :param audit_type: Comma-separated list of audit types
        :type audit_type: str or list, optional
        :param sort_order: Sort order of the results
        :type sort_order: str, optional
        :return: list of audit objects
        :rtype: list
        """
        params = {'eventId': event_id,
                  'auditType': audit_type,
                  'sortOrder': sort_order,
                  'at': self.process_time_param(at_time),
                  'from': self.process_time_param(from_time),
                  'to': self.process_time_param(to_time),
                  'after': self.process_time_param(after_time),
                  'before': self.process_time_param(before_time),
                  'offset': offset,
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_audit'))
        data = self.con.get(url=url, params=params)
        return Pagination(self, data, xmatters.factories.AuditFactory, limit=limit) if data.get('data') else []

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class DevicesEndpoint(ApiBridge):
    """ Used to interact with '/devices' endpoint """
    _endpoints = {'get_devices': '/devices',
                  'get_device_by_id': '/devices/{device_id}'}

    def __init__(self, parent):
        """
        :param parent: XMSession instance
        :type parent: :class:`xmatters.session.XMSession`
        """
        super(DevicesEndpoint, self).__init__(parent)

    def get_devices(self, device_status=None, device_type=None, device_names=None, phone_number_format=None,
                    offset=None, limit=None):
        params = {'deviceStatus': device_status,
                  'deviceType': device_type,
                  'phoneNumberFormat': phone_number_format,
                  'deviceNames': device_names,
                  'offset': offset,
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_devices'))
        data = self.con.get(url=url, params=params)
        return Pagination(self, data, xmatters.factories.DeviceFactory, limit=limit) if data.get('data') else []

    def get_device_by_id(self, device_id, at_time=None):
        params = {'at': self.process_time_param(at_time)}
        url = self.build_url(self._endpoints.get('get_device_by_id').format(device_id=device_id))
        data = self.con.get(url=url, params=params)
        return xmatters.factories.DeviceFactory.compose(self, data) if data else None

    def create_device(self, data):
        url = self.build_url(self._endpoints.get('get_devices'))
        data = self.con.post(url, data=data)
        return xmatters.factories.DeviceFactory.compose(self, data) if data else None

    def update_device(self, data):
        url = self.build_url(self._endpoints.get('get_devices'))
        data = self.con.post(url=url, data=data)
        return xmatters.factories.DeviceFactory.compose(self, data) if data else None

    def delete_device(self, device_id):
        url = self.build_url(self._endpoints.get('get_device_by_id').format(device_id=device_id))
        data = self.con.delete(url=url)
        return xmatters.factories.DeviceFactory.compose(self, data) if data else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class DeviceNamesEndpoint(ApiBridge):
    _endpoints = {'get_device_names': '/device-names',
                  'delete_device_name': '/device-names/{device_name_id}'}

    def __init__(self, parent):
        super(DeviceNamesEndpoint, self).__init__(parent)

    # TODO: Test params
    def get_device_names(self, device_types=None, search=None, sort_by=None, sort_order=None,
                         at_time=None, offset=None, limit=None):
        params = {'search': ' '.join(search) if search else None,
                  'sortBy': sort_by,
                  'sortOrder': sort_order,
                  'deviceTypes': device_types,
                  'at': self.process_time_param(at_time),
                  'offset': offset,
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_device_names'))
        data = self.con.get(url=url, params=params)
        return Pagination(self, data, factory.DeviceNameFactory, limit=limit) if data.get('data') else []

    def create_device_name(self, data):
        url = self.build_url(self._endpoints.get('get_device_names'))
        data = self.con.post(url, data=data)
        return xmatters.factories.DeviceNameFactory.compose(self, data) if data else None

    def update_device_name(self, data):
        url = self.build_url(self._endpoints.get('get_device_names'))
        data = self.con.post(url, data=data)
        return xmatters.factories.DeviceNameFactory.compose(self, data) if data else None

    def delete_device_name(self, device_name_id):
        url = self.build_url(self._endpoints.get('delete_device_name').format(device_name_id=device_name_id))
        data = self.con.delete(url)
        return xmatters.factories.DeviceNameFactory.compose(self, data) if data else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class DeviceTypesEndpoint(ApiBridge):
    _endpoints = {'get_device_types': '/device-types'}

    def __init__(self, parent):
        super(DeviceTypesEndpoint, self).__init__(parent)

    def get_device_types(self):
        url = self.build_url(self._endpoints.get('get_device_types'))
        data = self.con.get(url)
        return DeviceTypes(data) if data else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class DynamicTeamsEndpoint(ApiBridge):
    _endpoints = {'get_dynamic_teams': '/dynamic-teams',
                  'get_dynamic_team_by_id': '/dynamic-teams/{dynamic_team_id}'}

    # TODO: Test params
    def get_dynamic_teams(self, search=None, operand=None, fields=None, supervisors=None, sort_by=None,
                          sort_order=None, offset=None, limit=None):
        params = {'search': ' '.join(search) if search else None,
                  'operand': operand,
                  'fields': fields,
                  'supervisors': supervisors,
                  'sortBy': sort_by,
                  'sortOrder': sort_order,
                  'offset': offset,
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_dynamic_teams'))
        data = self.con.get(url, params)
        return Pagination(self, data, DynamicTeam, limit=limit) if data.get('data') else []

    def get_dynamic_team_by_id(self, dynamic_team_id):
        url = self.build_url(self._endpoints.get('get_dynamic_team_by_id').format(dynamic_team_id=dynamic_team_id))
        data = self.con.get(url)
        return DynamicTeam(self, data) if data else None

    # TODO: Test
    def create_dynamic_team(self, data):
        url = self.build_url(self._endpoints.get('get_dynamic_teams'))
        data = self.con.post(url, data=data)
        return DynamicTeam(self, data) if data else None

    # TODO: Test
    def update_dynamic_team(self, data):
        url = self.build_url(self._endpoints.get('get_dynamic_teams'))
        data = self.con.post(url, data=data)
        return DynamicTeam(self, data) if data else None

    # TODO: Test
    def delete_dynamic_team(self, dynamic_team_id):
        url = self.build_url(self._endpoints.get('get_dynamic_team_by_id').format(dynamic_team_id=dynamic_team_id))
        data = self.con.delete(url=url)
        return DynamicTeam(self, data) if data else None

    def __init__(self, parent):
        super(DynamicTeamsEndpoint, self).__init__(parent)

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class EventsEndpoint(ApiBridge):
    _endpoints = {'get_events': '/events',
                  'get_event_by_id': '/events/{event_id}',
                  'trigger_event': '{instance_url}/api/integration/1/functions/{func_id}/triggers'}

    def __init__(self, parent):
        super(EventsEndpoint, self).__init__(parent)

    # TODO: Test params
    def get_events(self, property_name=None, property_value=None, property_value_operator=None, status=None,
                   priority=None, plan=None, form=None, request_id=None, event_type=None, sort_by=None, sort_order=None,
                   submitter_id=None, search=None, targeted_recipients=None, resolved_users=None, from_time=None,
                   to_time=None, at_time=None, after_time=None, before_time=None, offset=None, limit=None):
        params = {'propertyName': property_name,
                  'propertyValue': property_value,
                  'propertyValueOperator': property_value_operator,
                  'status': status,
                  'priority': priority,
                  'plan': plan,
                  'form': form,
                  'requestId': request_id,
                  'eventType': event_type,
                  'sortBy': sort_by,
                  'sortOrder': sort_order,
                  'submitterid': submitter_id,
                  'search': ' '.join(search) if search else None,
                  'targetedRecipients': targeted_recipients,
                  'resolvedUsers': resolved_users,
                  'from': self.process_time_param(from_time),
                  'to': self.process_time_param(to_time),
                  'at': self.process_time_param(at_time),
                  'after': self.process_time_param(after_time),
                  'before': self.process_time_param(before_time),
                  'offset': offset,
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_events'))
        data = self.con.get(url, params)
        return Pagination(self, data, Event, limit=limit) if data.get('data') else []

    # TODO: Test params
    def get_event_by_id(self, event_id, at=None):
        params = {'at': self.process_time_param(at)}
        url = self.build_url(self._endpoints.get('get_event_by_id').format(event_id=event_id))
        data = self.con.get(url, params)
        return Event(self, data) if data else None

    # TODO
    # def trigger_event(self, function_id, data, params=None):
    #     url = self._endpoints.get('trigger_event').format(instance_url=self.con.instance_url, func_id=function_id)
    #     data = self.con.post(url, data=data, params=params)
    #     return RequestReference(data) if data else None

    # TODO: Test
    def change_event_status(self, data):
        url = self.build_url(self._endpoints.get('get_events'))
        data = self.con.post(url, data=data)
        return Event(self, data) if data else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class EventSuppressionsEndpoint(ApiBridge):
    _endpoints = {'get_event_suppressions_by_event_id': '/event-suppressions?event={event_id}'}

    def __init__(self, parent):
        super(EventSuppressionsEndpoint, self).__init__(parent)

    def get_suppressions_by_event_id(self, event_id, sort_by, sort_order, offset=None, limit=None):
        params = {'sortBy': sort_by,
                  'sortOrder': sort_order,
                  'offset': offset,
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_event_suppressions_by_event_id').format(event_id))
        data = self.con.get(url, params)
        return Pagination(self, data, EventSuppression, limit=limit) if data.get('data') else []

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class ConferenceBridgesEndpoint(ApiBridge):
    _endpoints = {'get_conference_bridges': '/conference-bridges',
                  'get_conference_bridge_by_id': '/conference-bridges/{bridge_id}'}

    def __init__(self, parent):
        super(ConferenceBridgesEndpoint, self).__init__(parent)

    # TODO: Test params
    def get_conference_bridges(self, name=None, description=None, toll_number=None, toll_free_number=None,
                               preferred_connection_type=None, pause_before_bridge_prompt=None,
                               static_bridge_number=False, bridge_number=None, dial_after_bridge=None, offset=None,
                               limit=None):
        params = {'name': name,
                  'description': description,
                  'tollNumber': toll_number,
                  'tollFreeNumber': toll_free_number,
                  'preferredConnectionType': preferred_connection_type,
                  'pauseBeforeBridgePrompt': pause_before_bridge_prompt,
                  'staticBridgeNumber': static_bridge_number,
                  'bridgeNumber': bridge_number,
                  'dialAfterBridge': dial_after_bridge,
                  'offset': offset,
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_conference_bridges'))
        data = self.con.get(url, params)
        return Pagination(self, data, ConferenceBridge, limit=limit) if data.get('data') else []

    def get_conference_bridge_by_id(self, bridge_id):
        url = self.build_url(self._endpoints.get('get_conference_bridge_by_id').format(bridge_id=bridge_id))
        data = self.con.get(url)
        return ConferenceBridge(self, data) if data else None

    def create_conference_bridge(self, data):
        url = self.build_url(self._endpoints.get('get_conference_bridges'))
        data = self.con.post(url, data=data)
        return ConferenceBridge(self, data) if data else None

    def update_conference_bridge(self, data):
        url = self.build_url(self._endpoints.get('get_conference_bridges'))
        data = self.con.post(url, data=data)
        return ConferenceBridge(self, data) if data else None

    def delete_conference_bridge(self, bridge_id):
        url = self.build_url(self._endpoints.get('get_conference_bridge_by_id').format(bridge_id=bridge_id))
        data = self.con.delete(url=url)
        return ConferenceBridge(self, data) if data else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class FormsEndpoint(ApiBridge):
    _endpoints = {'get_forms': '/forms',
                  'get_form_by_id': '/forms/{form_id}'}

    def __init__(self, parent):
        super(FormsEndpoint, self).__init__(parent)

    # TODO: Test params
    def get_forms(self, search=None, operand=None, fields=None, enabled_for=None, plan_type=None, sort_by=None,
                  sort_order=None, trigger_type=None, offset=None, limit=None):
        params = {'search': ' '.join(search) if search else None,
                  'operand': operand,
                  'fields': fields,
                  'enabledFor': enabled_for,
                  'plans.planType': plan_type,
                  'sortBy': sort_by,
                  'sortOrder': sort_order,
                  'triggerType': trigger_type,
                  'offset': offset,
                  'limit': limit}

        url = self.build_url(self._endpoints.get('get_forms'))
        data = self.con.get(url, params=params)
        return Pagination(self, data, xmatters.xm_objects.forms.Form, limit=limit) if data.get('data') else []

    # TODO: Test
    def get_form_by_id(self, form_id):
        url = self.build_url(self._endpoints.get('get_form_by_id').format(form_id=form_id))
        data = self.con.get(url)
        return xmatters.xm_objects.forms.Form(self, data) if data else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class GroupsEndpoint(ApiBridge):
    _endpoints = {'get_groups': '/groups',
                  'get_group_by_id': '/groups/{group_id}'}

    def __init__(self, parent):
        super(GroupsEndpoint, self).__init__(parent)

    # TODO: Test params
    def get_groups(self, search=None, operand=None, fields=None, sites=None, members=None, members_exists=None,
                   sort_by=None, sort_order=None, status=None, supervisors=None, offset=None, limit=None):
        params = {'search': ' '.join(search) if search else None,
                  'operand': operand,
                  'fields': fields,
                  'sites': sites,
                  'members': members,
                  'members.exists': members_exists,
                  'sortBy': sort_by,
                  'sortOrder': sort_order,
                  'status': status,
                  'supervisors': supervisors,
                  'offset': offset,
                  'limit': limit}

        url = self.build_url(self._endpoints.get('get_groups'))
        data = self.con.get(url, params=params)
        return Pagination(self, data, Group) if data.get('data') else []

    def get_group_by_id(self, group_id, at=None):
        params = {'at': self.process_time_param(at)}
        url = self.build_url(self._endpoints.get('get_group_by_id').format(group_id=group_id))
        data = self.con.get(url, params=params)
        return Group(self, data) if data else None

    # TODO: Test
    def create_group(self, data):
        url = self.build_url(self._endpoints.get('get_groups'))
        data = self.con.post(url, data=data)
        return Group(self, data) if data else None

    # TODO: Test
    def update_group(self, data):
        url = self.build_url(self._endpoints.get('get_groups'))
        data = self.con.post(url, data=data)
        return Group(self, data) if data else None

    # TODO: Test
    def delete_group(self, group_id):
        url = self.build_url(self._endpoints.get('get_group_by_id').format(group_id=group_id))
        data = self.con.delete(url)
        return Group(self, data) if data else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class ImportJobsEndpoint(ApiBridge):
    _endpoints = {'get_import_jobs': '/imports',
                  'get_import_job_by_id': '/imports/{import_id}'}

    def __init__(self, parent):
        super(ImportJobsEndpoint, self).__init__(parent)

    # TODO: Test params
    def get_import_jobs(self, transform_type=None, sort_by=None, sort_order=None):
        params = {'transformType': transform_type,
                  'sortBy': sort_by,
                  'sortOrder': sort_order}

        url = self.build_url(self._endpoints.get('get_import_jobs'))
        data = self.con.get(url, params=params).get('data', {})
        return [Import(self, job) for job in data] if data else []

    # TODO: Test
    def get_import_job_by_id(self, import_id):
        url = self.build_url(self._endpoints.get('get_import_job_by_id').format(import_id=import_id))
        data = self.con.get(url)
        return Import(self, data) if data else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class IncidentsEndpoint(ApiBridge):
    _endpoints = {'get_incidents': '/incidents',
                  'get_incident_by_id': '/incidents/{incident_id}'}

    def __init__(self, parent):
        super(IncidentsEndpoint, self).__init__(parent)

    def get_incidents(self, request_id=None, search=None, operand=None, fields=None, sites=None, status=None,
                      severity=None, offset=None, limit=None):
        params = {'requestId': request_id,
                  'search': ' '.join(search) if search else None,
                  'operand': operand,
                  'fields': fields,
                  'sites': sites,
                  'status': status,
                  'severity': severity,
                  'offset': offset,
                  'limit': limit}

        url = self.build_url(self._endpoints.get('get_incidents'))
        data = self.con.get(url, params=params)
        return Pagination(self, data, Incident, limit=limit) if data.get('data') else []

    # TODO: Test
    def get_incident_by_id(self, incident_id):
        url = self.build_url(self._endpoints.get('get_incident_by_id').format(incident_id=incident_id))
        data = self.con.get(url)
        return Incident(self, data) if data else None

    # TODO
    # def trigger_incident(self, data, params=None):
    #     url = self._endpoints.get('trigger_incident').format(instance_url=self.con.instance_url)
    #     data = self.con.post(url, data=data, params=params)
    #     return RequestReference(data) if data else None

    # TODO: Test
    def update_incident(self, incident_id, data):
        url = self.build_url(self._endpoints.get('get_incident_by_id').format(incident_id=incident_id))
        data = self.con.post(url, data=data)
        return Incident(self, data) if data else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class OnCallEndpoint(ApiBridge):
    _endpoints = {'get_oncall': '/on-call'}

    def __init__(self, parent):
        super(OnCallEndpoint, self).__init__(parent)

    # TODO: Test params
    def get_oncall(self, groups, members_per_shift=None, at_time=None, from_time=None, to_time=None, offset=None,
                   limit=None):
        params = {'groups': groups,
                  'membersPerShift': members_per_shift,
                  'at': self.process_time_param(at_time),
                  'from': self.process_time_param(from_time),
                  'to': self.process_time_param(to_time),
                  'offset': offset,
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_oncall'))
        data = self.con.get(url, params=params)
        return Pagination(self, data, OnCall, limit=limit) if data.get('data') else []

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class OnCallSummaryEndpoint(ApiBridge):
    _endpoints = {'get_oncall_summary': '/on-call-summary'}

    def __init__(self, parent):
        super(OnCallSummaryEndpoint, self).__init__(parent)

    # TODO: Test params
    def get_oncall_summary(self, groups, escalation_order=None, recipients_per_shift=None, at=None):
        params = {'groups': groups,
                  'escalationOrder': escalation_order,
                  'recipientsPerShift': recipients_per_shift,
                  'at': self.process_time_param(at)}

        url = self.build_url(self._endpoints.get('get_oncall_summary'))
        data = self.con.get(url, params=params)
        return [OnCallSummary(self, summary) for summary in data] if data else []

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class PeopleEndpoint(ApiBridge):
    _endpoints = {'get_people': '/people',
                  'get_person_by_id': '/people/{person_id}'}

    def __init__(self, parent):
        super(PeopleEndpoint, self).__init__(parent)

    # TODO: Test params
    def get_people(self, search=None, operand=None, fields=None, property_names=None, property_values=None,
                   devices_exist=None, devices_test_status=None, site=None, status=None, supervisors_exists=None,
                   groups=None, groups_exist=None, roles=None, supervisors=None, created_from_time=None,
                   created_to_time=None, created_before_time=None, created_after_time=None, sort_by=None,
                   sort_order=None, at_time=None, limit=None):
        params = {'search': ' '.join(search) if search else None,
                  'operand': operand,
                  'fields': fields,
                  'propertyNames': property_names,
                  'propertyValues': property_values,
                  'devices.exists': devices_exist,
                  'devices.testStatus': devices_test_status,
                  'site': site,
                  'status': status,
                  'supervisors.exists': supervisors_exists,
                  'groups': groups,
                  'groups.exists': groups_exist,
                  'roles': roles,
                  'supervisors': supervisors,
                  'createdFrom': self.process_time_param(created_from_time),
                  'createdTo': self.process_time_param(created_to_time),
                  'createdBefore': self.process_time_param(created_before_time),
                  'createdAfter': self.process_time_param(created_after_time),
                  'sortBy': sort_by,
                  'sortOrder': sort_order,
                  'at': self.process_time_param(at_time),
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_people'))
        data = self.con.get(url, params=params)
        return Pagination(self, data, Person, limit=limit) if data.get('data') else []

    def get_person_by_id(self, person_id, params=None):
        url = self.build_url(self._endpoints.get('get_person_by_id').format(person_id=person_id))
        data = self.con.get(url, params)
        return Person(self, data) if data else None

    def get_people_by_query(self, first_name=None, last_name=None, target_name=None, web_login=None, phone_number=None,
                            email_address=None, offset=None, limit=None):
        if all(p is None for p in (first_name, last_name, target_name, web_login, phone_number, email_address)):
            raise ValueError('must assign a parameter to query by')
        params = {'firstName': first_name,
                  'lastName': last_name,
                  'targetName': target_name,
                  'webLogin': web_login,
                  'phoneNumber': phone_number,
                  'emailAddress': email_address,
                  'offset': offset,
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_people'))
        data = self.con.get(url, params=params)
        return Pagination(self, data, Person, limit=limit) if data.get('data') else []

    def create_person(self, data):
        url = self.build_url(self._endpoints.get('get_people'))
        data = self.con.post(url, data=data)
        return Person(self, data) if data else None

    def update_person(self, data):
        url = self.build_url(self._endpoints.get('get_people'))
        data = self.con.post(url, data=data)
        return Person(self, data) if data else None

    def delete_person(self, person_id):
        url = self.build_url(self._endpoints.get('get_person_by_id').format(person_id=person_id))
        data = self.con.delete(url)
        return Person(self, data) if data else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class PlansEndpoint(ApiBridge):
    _endpoints = {'get_plans': '/plans',
                  'get_plan_by_id': '/plans/{plan_id}'}

    def __init__(self, parent):
        super(PlansEndpoint, self).__init__(parent)

    # TODO: Test params
    def get_plans(self, plan_type=None, enabled=None, enabled_for=None, search=None, operand=None, fields=None,
                  sort_by=None, sort_order=None, subscription_forms=None, at=None, offset=None, limit=None):
        params = {'planType': plan_type,
                  'enabled': enabled,
                  'enabledFor': enabled_for,
                  'search': ' '.join(search) if search else None,
                  'operand': operand,
                  'fields': fields,
                  'sortBy': sort_by,
                  'sortOrder': sort_order,
                  'subscription-forms': subscription_forms,
                  'at': self.process_time_param(at),
                  'offset': offset,
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_plans'))
        data = self.con.get(url, params=params)
        return Pagination(self, data, Plan, limit=limit) if data.get('data') else []

    def get_plan_by_id(self, plan_id, at=None):
        params = {'at': self.process_time_param(at)}
        url = self.build_url(self._endpoints.get('get_plan_by_id').format(person_id=plan_id))
        data = self.con.get(url, params=params)
        return Plan(self, data) if data else None

    # TODO: Test
    def create_plan(self, data):
        url = self.build_url(self._endpoints.get('get_plans'))
        data = self.con.post(url, data=data)
        return Plan(self, data) if data else None

    # TODO: Test
    def update_plan(self, data):
        url = self.build_url(self._endpoints.get('get_plans'))
        data = self.con.post(url, data=data)
        return Plan(self, data) if data else None

    # TODO: Test
    def delete_plan(self, plan_id):
        url = self.build_url(self._endpoints.get('get_plan_by_id').format(plan_id=plan_id))
        data = self.con.delete(url)
        return Plan(self, data) if data else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class RolesEndpoint(ApiBridge):
    _endpoints = {'get_roles': '/roles'}

    def __init__(self, parent):
        super(RolesEndpoint, self).__init__(parent)

    # TODO: Test params
    def get_roles(self, name=None, allow_wildcards=None, offset=None, limit=None):
        params = {'name': name,
                  'allowWildcards': allow_wildcards,
                  'offset': offset,
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_roles'))
        data = self.con.get(url, params=params)
        return Pagination(self, data, Role, limit=limit) if data.get('data') else []

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class ScenariosEndpoint(ApiBridge):
    _endpoints = {'get_scenarios': '/scenarios',
                  'get_scenario_by_id': '/scenarios/{scenario_id}'}

    def __init__(self, parent):
        super(ScenariosEndpoint, self).__init__(parent)

    # TODO: Test params
    def get_scenarios(self, search=None, operand=None, enabled_for=None, offset=None, limit=None):
        params = {'search': ' '.join(search) if search else None,
                  'operand': operand,
                  'enabledFor': enabled_for,
                  'offset': offset,
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_scenarios'))
        data = self.con.get(url, params=params)
        return Pagination(self, data, Scenario, limit=limit) if data.get('data') else []

    def get_scenario_by_id(self, scenario_id):
        url = self.build_url(self._endpoints.get('get_scenario_by_id').format(scenario_id=scenario_id))
        data = self.con.get(url)
        return Scenario(self, data) if data else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class ServicesEndpoint(ApiBridge):
    _endpoints = {'get_services': '/services',
                  'get_service_by_id': '/scenarios/{service_id}'}

    def __init__(self, parent):
        super(ServicesEndpoint, self).__init__(parent)

    # TODO: Test params
    def get_services(self, search=None, operand=None, fields=None, owned_by=None, offset=None, limit=None):
        params = {'search': ' '.join(search) if search else None,
                  'operand': operand,
                  'fields': fields,
                  'ownedBy': owned_by,
                  'offset': offset,
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_services'))
        data = self.con.get(url, params)
        return Pagination(self, data, Service, limit=limit) if data.get('data') else []

    def get_service_by_id(self, service_id):
        url = self.build_url(self._endpoints.get('get_service_by_id').format(service_id=service_id))
        data = self.con.get(url)
        return Service(self, data) if data else None

    # TODO: Test
    def create_service(self, data):
        url = self.build_url(self._endpoints.get('get_services'))
        data = self.con.post(url, data=data)
        return Service(self, data) if data else None

    # TODO: Test
    def update_service(self, data):
        url = self.build_url(self._endpoints.get('get_services'))
        data = self.con.post(url, data=data)
        return Service(self, data) if data else None

    # TODO: Test
    def delete_service(self, service_id):
        url = self.build_url(self._endpoints.get('get_service_by_id').format(service_id=service_id))
        data = self.con.delete(url)
        return Service(self, data) if data else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class SitesEndpoint(ApiBridge):
    _endpoints = {'get_sites': '/sites',
                  'get_site_by_id': '/sites/{site_id}'}

    def __init__(self, parent):
        super(SitesEndpoint, self).__init__(parent)

    # TODO: Test params
    def get_sites(self, search=None, operand=None, fields=None, sort_by=None, sort_order=None, country=None,
                  geocoded=None, status=None, offset=None, limit=None):
        params = {'search': ' '.join(search) if search else None,
                  'operand': operand,
                  'fields': fields,
                  'sortBy': sort_by,
                  'sortOrder': sort_order,
                  'country': country,
                  'geocoded': geocoded,
                  'status': status,
                  'offset': offset,
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_sites'))
        data = self.con.get(url, params)
        return Pagination(self, data, Site, limit=limit) if data.get('data') else []

    def get_site_by_id(self, site_id, params=None):
        url = self.build_url(self._endpoints.get('get_site_by_id').format(site_id=site_id))
        data = self.con.get(url, params)
        return Site(self, data) if data else None

    # TODO: Test
    def create_site(self, data):
        url = self.build_url(self._endpoints.get('get_sites'))
        data = self.con.post(url, data=data)
        return Site(self, data) if data else None

    # TODO: Test
    def update_site(self, data):
        url = self.build_url(self._endpoints.get('get_sites'))
        data = self.con.post(url, data=data)
        return Site(self, data) if data else None

    # TODO: Test
    def delete_site(self, site_id):
        url = self.build_url(self._endpoints.get('get_site_by_id').format(site_id=site_id))
        data = self.con.delete(url)
        return Site(self, data) if data else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class SubscriptionsEndpoint(ApiBridge):
    _endpoints = {'get_subscriptions': '/subscriptions',
                  'get_subscription_by_id': '/subscription-forms/{sub_id}',
                  'get_subscribers': '/subscribers',
                  'unsubscribe_person': '/subscribers/{person_id}'}

    def __init__(self, parent):
        super(SubscriptionsEndpoint, self).__init__(parent)

    # TODO: Test params
    def get_subscriptions(self, owner=None, subscriber=None, shared_with=None, managed_by=None, subscription_name=None,
                          subscription_description=None, subscription_form=None, offset=None, limit=None):
        params = {'owner': owner,
                  'subscriber': subscriber,
                  'sharedWith': shared_with,
                  'managedBy': managed_by,
                  'subscriptionName': subscription_name,
                  'subscriptionDescription': subscription_description,
                  'subscriptionForm': subscription_form,
                  'offset': offset,
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_subscriptions'))
        data = self.con.get(url, params=params)
        return Pagination(self, data, Subscription, limit=limit) if data else []

    def get_subscription_by_id(self, subscription_id):
        url = self.build_url(self._endpoints.get('get_subscription_by_id').format(sub_id=subscription_id))
        data = self.con.get(url)
        return SubscriptionForm(self, data) if data else None

    # TODO: Test
    def get_subscribers(self, subscription_id, offset=None, limit=None):
        params = {'id': subscription_id,
                  'offset': offset,
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_subscribers'))
        subscribers = self.con.get(url, params=params)
        return Pagination(self, subscribers, Person, limit=limit) if subscribers.get('data') else []

    # TODO: Test
    def unsubscribe_person(self, person_id):
        url = self.build_url(self._endpoints.get('unsubscribe_person').format(person_id=person_id))
        data = self.con.delete(url)
        return Subscription(self, data) if data else None

    # TODO: Test
    def create_subscription(self, data):
        url = self.build_url(self._endpoints.get('get_subscriptions'))
        data = self.con.post(url, data=data)
        return Subscription(self, data) if data else None

    # TODO: Test
    def update_subscription(self, data):
        url = self.build_url(self._endpoints.get('get_subscriptions'))
        data = self.con.post(url, data=data)
        return Subscription(self, data) if data else None

    # TODO: Test
    def delete_subscription(self, subscription_id):
        url = self.build_url(self._endpoints.get('get_subscription_by_id').format(sub_id=subscription_id))
        data = self.con.delete(url)
        return Subscription(self, data) if data else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class SubscriptionFormsEndpoint(ApiBridge):
    _endpoints = {'get_subscription_forms': '/subscription-forms',
                  'get_subscription_form_id': '/subscription-forms/{sub_form_id}'}

    def __init__(self, parent):
        super(SubscriptionFormsEndpoint, self).__init__(parent)

    # TODO: Test params
    def get_subscription_forms(self, sort_by=None, sort_order=None, offset=None, limit=None):
        params = {'sortBy': sort_by,
                  'sortOrder': sort_order,
                  'offset': offset,
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_subscription_forms'))
        data = self.con.get(url, params=params)
        return Pagination(self, data, SubscriptionForm, limit=limit) if data.get('data') else []

    def get_subscription_form_by_id(self, sub_form_id):
        url = self.build_url(self._endpoints.get('get_subscription_form_by_id').format(sub_form_id=sub_form_id))
        data = self.con.get(url)
        return SubscriptionForm(self, data) if data else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class TemporaryAbsencesEndpoint(ApiBridge):
    _endpoints = {'get_temporary_absences': '/temporary-absences',
                  'delete_temporary_absence': '/temporary-absences/{temp_abs_id}'}

    def __init__(self, parent):
        super(TemporaryAbsencesEndpoint, self).__init__(parent)

    # TODO: Test params
    def get_temporary_absences(self, member=None, groups=None, absence_type=None, from_time=None, to_time=None,
                               offset=None, limit=None):
        params = {'member': member,
                  'groups': groups,
                  'absenceType': absence_type,
                  'from': self.process_time_param(from_time),
                  'to': self.process_time_param(to_time),
                  'offset': offset,
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_temporary_absences'))
        data = self.con.get(url, params)
        return Pagination(self, data, TemporaryAbsence, limit=limit) if data.get('data') else []

    # TODO: Test
    def create_temporary_absence(self, data):
        url = self.build_url(self._endpoints.get('get_temporary_absences'))
        data = self.con.post(url, data=data)
        return TemporaryAbsence(self, data) if data else None

    # TODO: Test
    def delete_temporary_absence(self, temporary_absence_id):
        url = self.build_url(self._endpoints.get('delete_temporary_absence').format(temp_abs_id=temporary_absence_id))
        data = self.con.delete(url)
        return TemporaryAbsence(self, data) if data else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


# TODO
class UploadUsersEndpoint(ApiBridge):
    _endpoints = {'upload_user_upload_file': '/uploads/users-v1',
                  'upload_epic_zipsync_file': '/uploads/epic-v1'}

    def __init__(self, parent):
        super(UploadUsersEndpoint, self).__init__(parent)

    def upload_user_upload_file(self, file_path):
        pass

    def upload_epic_zipsync_file(self, file_path):
        pass

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()



import xmatters.utils as util
from xmatters.xm_objects.common import Recipient
from xmatters.xm_objects.people import PersonReference
from xmatters.connection import ApiBridge
import xmatters.xm_objects.events as events


class Notification(ApiBridge):
    def __init__(self, parent, data):
        super(Notification, self).__init__(parent, data)
        self.id = data.get('id')
        self.category = data.get('category')
        recipient = data.get('recipient')
        self.recipient = Recipient(parent, recipient) if recipient else None
        self.delivery_status = data.get('deliveryStatus')
        created = data.get('created')
        self.created = util.TimeAttribute(created) if created else None
        event = data.get('event')
        self.event = events.EventReference(parent, event) if event else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class Response(ApiBridge):
    def __init__(self, parent, data):
        super(Response, self).__init__(parent, data)
        self.comment = data.get('comment')
        notification = data.get('notification')
        self.notification = Notification(self, notification) if notification else None
        options = data.get('options', {}).get('data')
        self.options = [events.ResponseOption(r) for r in options] if options else None
        self.source = data.get('source')
        received = data.get('received')
        self.received = util.TimeAttribute(received) if received else None
        self.response = data.get('response')

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class AuditBase(ApiBridge):
    def __init__(self, parent, data):
        super(AuditBase, self).__init__(parent, data)
        self.id = data.get('id')
        self.type = data.get('type')
        self.order_id = data.get('orderId')
        at = data.get('at')
        self.at = util.TimeAttribute(at) if at else None

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.type)

    def __str__(self):
        return self.__repr__()


class Audit(AuditBase):
    def __init__(self, parent, data):
        super(Audit, self).__init__(parent, data)

        event = data.get('event')
        self.event = events.EventReference(parent, event) if event else None

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.type)

    def __str__(self):
        return self.__repr__()


class Annotation(ApiBridge):
    def __init__(self, parent, data):
        super(Annotation, self).__init__(parent, data)
        event = data.get('event')
        self.event = events.EventReference(parent, event) if event else None
        author = data.get('author')
        self.author = PersonReference(parent, author) if author else None
        self.comment = data.get('comment')

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class AuditNotification(AuditBase):
    def __init__(self, parent, data):
        super(AuditNotification, self).__init__(parent, data)
        notification = data.get('notification')
        self.notification = Notification(self, notification) if notification else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class AuditAnnotation(AuditBase):
    def __init__(self, parent, data):
        super(AuditAnnotation, self).__init__(parent, data)
        annotation = data.get('annotation')
        self.annotation = Annotation(self, annotation) if annotation else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class AuditResponse(AuditBase):
    def __init__(self, parent, data):
        super(AuditResponse, self).__init__(parent, data)
        response = data.get('response')
        self.response = Response(self, response) if response else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()



import xmatters.xm_objects.common
import xmatters.xm_objects.plans
import xmatters.connection


class PlanConstant(xmatters.connection.ApiBridge):
    def __init__(self, parent, data):
        super(PlanConstant, self).__init__(parent, data)
        self.id = data.get('id')
        plan = data.get('plan')
        self.plan = xmatters.xm_objects.plans.PlanPointer(plan) if plan else None
        self.name = data.get('name')
        self.value = data.get('value')
        self.description = data.get('description')
        links = data.get('links')
        self.links = xmatters.xm_objects.common.SelfLink(self, links) if links else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()



class TargetDeviceNameSelector(object):
    def __init__(self, data):
        self.name = data.get('name')
        self.selected = data.get('selected')
        self.visible = data.get('visible')

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.name)

    def __str__(self):
        return self.__repr__()


class DeviceName(object):
    def __init__(self, data):
        self.id = data.get('id')
        self.device_type = data.get('deviceType')
        self.name = data.get('name')
        self.description = data.get('description')
        self.privileged = data.get('privileged')

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.name)

    def __str__(self):
        return self.__repr__()


class DeviceNameEmail(DeviceName):
    def __init__(self, data):
        super(DeviceNameEmail, self).__init__(data)
        self.domains = data.get('domains', [])

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.name)

    def __str__(self):
        return self.__repr__()




import xmatters.factories as factory
import xmatters.xm_objects.forms
from xmatters.xm_objects.plan_endpoints import Endpoint
from xmatters.connection import ApiBridge
from xmatters.xm_objects.common import Pagination, SelfLink
from xmatters.xm_objects.integrations import Integration
from xmatters.xm_objects.people import Person
from xmatters.xm_objects.plan_constants import PlanConstant
from xmatters.xm_objects.shared_libraries import SharedLibrary
from xmatters.xm_objects.subscription_forms import SubscriptionForm


class PlanPointer(object):
    def __init__(self, data):
        self.id = data.get('id')

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class PlanReference(object):
    def __init__(self, data):
        self.id = data.get('id')
        self.name = data.get('name')

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class Plan(ApiBridge):
    _endpoints = {'creator': '?embed=creator',
                  'get_forms': '/forms',
                  'get_form_by_id': '/forms/{form_id}',
                  'get_integrations': '/integrations',
                  'get_integration_by_id': '/integrations/{int_id}',
                  'get_constants': '/constants',
                  'get_properties': '/property-definitions',
                  'get_libraries': '/shared-libraries',
                  'get_library_by_id': '/shared-libraries/{lib_id}',
                  'get_endpoints': '/endpoints',
                  'get_subscription_forms': '/subscription-forms',
                  'delete_constant': '/constants/{const_id}',
                  'delete_endpoints': '/endpoints/{end_id}',
                  'delete_property': '/property-definitions/{prop_id}'}

    def __init__(self, parent, data):
        super(Plan, self).__init__(parent, data)
        self.id = data.get('id')
        self.plan_type = data.get('planType')
        self.name = data.get('name')
        self.description = data.get('description')
        self.enabled = data.get('enabled')
        self.editable = data.get('editable')
        self.logging_level = data.get('loggingLevel')
        self.accessible = data.get('accessibleByAll')
        self.flood_control = data.get('floodControl')
        self.created = data.get('created')
        links = data.get('links')
        self.links = SelfLink(self, links) if links else None
        self.position = data.get('position')

    def get_forms(self, enabled_for=None, sort_by=None, sort_order=None, trigger_type=None):

        params = {'enabledFor': enabled_for,
                  'sortBy': sort_by,
                  'sortOrder': sort_order,
                  'triggerType': trigger_type}
        url = self.build_url(self._endpoints.get('get_forms'))
        fs = self.con.get(url, params=params)
        return Pagination(self, fs, xmatters.xm_objects.forms.Form) if fs.get('data') else []

    # TODO: Test
    def get_form_by_id(self, form_id, recipients=None):
        params = {'recipients': recipients}
        url = self.build_url(self._endpoints.get('get_form_by_id').format(form_id=form_id))
        data = self.con.get(url, params=params)
        return xmatters.xm_objects.forms.Form(self, data) if data else None

    # TODO: Test
    def create_form(self, data):
        url = self.build_url(self._endpoints.get('get_forms'))
        data = self.con.post(url, data=data)
        return xmatters.xm_objects.forms.Form(self, data) if data else None

    # TODO: Test
    def update_form(self, data):
        url = self.build_url(self._endpoints.get('get_forms'))
        data = self.con.post(url, data=data)
        return xmatters.xm_objects.forms.Form(self, data) if data else None

    def get_constants(self):
        url = self.build_url(self._endpoints.get('get_constants'))
        constants = self.con.get(url)
        return Pagination(self, constants, PlanConstant) if constants.get('data') else []

    # TODO: Test
    def create_constant(self, data):
        url = self.build_url(self._endpoints.get('get_constants'))
        data = self.con.post(url, data=data)
        return PlanConstant(self, data) if data else None

    # TODO: Test
    def update_constant(self, data):
        url = self.build_url(self._endpoints.get('get_constants'))
        data = self.con.post(url, data=data)
        return PlanConstant(self, data) if data else None

    # TODO: Test
    def delete_constant(self, constant_id):
        url = self.build_url(self._endpoints.get('delete_constant').format(const_id=constant_id))
        data = self.con.delete(url)
        return PlanConstant(self, data) if data else None

    def get_integrations(self, integration_type=None, deployed=None):
        params = {'integrationType': integration_type,
                  'deployed': deployed}
        url = self.build_url(self._endpoints.get('get_integrations'))
        ints = self.con.get(url, params=params)
        return Pagination(self, ints, Integration) if ints.get('data') else []

    # TODO: Test
    def get_integration_by_id(self, integration_id):
        url = self.build_url(self._endpoints.get('get_integration_by_id').format(int_id=integration_id))
        data = self.con.get(url)
        return Integration(self, data) if data else None

    # TODO: Test
    def create_integration(self, data):
        url = self.build_url(self._endpoints.get('get_integrations'))
        data = self.con.post(url, data=data)
        return Integration(self, data) if data else None

    # TODO: Test
    def update_integration(self, data):
        url = self.build_url(self._endpoints.get('get_integrations'))
        data = self.con.post(url, data=data)
        return Integration(self, data) if data else None

    # TODO: Test
    def delete_integration(self, integration_id):
        url = self.build_url(self._endpoints.get('get_integration_by_id').format(int_id=integration_id))
        data = self.con.delete(url)
        return Integration(self, data) if data else None

    def get_properties(self):
        url = self.build_url(self._endpoints.get('get_properties'))
        props = self.con.get(url)
        return Pagination(self, props, factory.PropertiesFactory) if props.get('data') else []

    # TODO: Test
    def create_property(self, data):
        url = self.build_url(self._endpoints.get('get_properties'))
        data = self.con.post(url, data=data)
        return factory.PropertiesFactory.compose(self, data) if data else None

    # TODO: Test
    def update_property(self, data):
        url = self.build_url(self._endpoints.get('get_properties'))
        data = self.con.post(url, data=data)
        return factory.PropertiesFactory.compose(self, data) if data else None

    # TODO: Test
    def delete_property(self, property_id):
        url = self.build_url(self._endpoints.get('delete_property').format(prop_id=property_id))
        data = self.con.delete(url)
        return factory.PropertiesFactory.compose(self, data) if data else None

    # TODO: Test param
    def get_shared_libraries(self, offset=None, limit=None):
        params = {'offset': offset,
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_libraries'))
        libs = self.con.get(url, params=params)
        return Pagination(self, libs, SharedLibrary, limit=limit) if libs.get('data') else []

    # TODO: Test
    def get_shared_library_by_id(self, library_id):
        url = self.build_url(self._endpoints.get('get_library_by_id').format(lib_id=library_id))
        data = self.con.get(url)
        return SharedLibrary(data) if data else None

    # TODO: Test
    def create_shared_library(self, data):
        url = self.build_url(self._endpoints.get('get_libraries'))
        data = self.con.post(url, data=data)
        return SharedLibrary(data) if data else None

    # TODO: Test
    def update_shared_library(self, data):
        url = self.build_url(self._endpoints.get('get_libraries'))
        data = self.con.post(url, data=data)
        return SharedLibrary(data) if data else None

    # TODO: Test
    def delete_shared_library(self, library_id):
        url = self.build_url(self._endpoints.get('get_library_by_id').format(lib_id=library_id))
        data = self.con.delete(url)
        return SharedLibrary(data) if data else None

    def get_endpoints(self):
        url = self.build_url(self._endpoints.get('get_endpoints'))
        endpoints = self.con.get(url)
        return Pagination(self, endpoints, xmatters.xm_objects.plan_endpoints.Endpoint) if endpoints.get('data') else []

    # TODO: Test
    def create_endpoint(self, data):
        url = self.build_url(self._endpoints.get('get_endpoints'))
        data = self.con.post(url, data=data)
        return xmatters.xm_objects.plan_endpoints.Endpoint(self, data) if data else None

    # TODO: Test
    def update_endpoint(self, data):
        url = self.build_url(self._endpoints.get('get_endpoints'))
        data = self.con.post(url, data=data)
        return xmatters.xm_objects.plan_endpoints.Endpoint(self, data) if data else None

    # TODO: Test
    def delete_endpoint(self, endpoint_id):
        url = self.build_url(self._endpoints.get('delete_endpoint').format(end_id=endpoint_id))
        data = self.con.delete(url)
        return xmatters.xm_objects.plan_endpoints.Endpoint(self, data) if data else None

    # TODO: Test params
    def get_subscription_forms(self, sort_by=None, sort_order=None, offset=None, limit=None):
        params = {'sortBy': sort_by,
                  'sortOrder': sort_order,
                  'offset': offset,
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_subscription_forms'))
        sub_forms = self.con.get(url, params=params)
        return Pagination(self, sub_forms, SubscriptionForm, limit=limit) if sub_forms.get('data') else []

    # TODO: Test
    def create_subscription_form(self, data):
        url = self.build_url(self._endpoints.get('get_subscription_forms'))
        data = self.con.post(url, data=data)
        return SubscriptionForm(self, data) if data else None

    # TODO: Test
    def update_subscription_form(self, data):
        url = self.build_url(self._endpoints.get('get_subscription_forms'))
        data = self.con.post(url, data=data)
        return SubscriptionForm(self, data) if data else None

    @property
    def creator(self):
        url = self.build_url(self._endpoints.get('creator'))
        creator = self.con.get(url).get('creator')
        return Person(self, creator) if creator else None

    @property
    def constants(self):
        return self.get_constants()

    @property
    def endpoints(self):
        return self.get_endpoints()

    @property
    def forms(self):
        return self.get_forms()

    @property
    def integrations(self):
        return self.get_integrations()

    @property
    def property_definitions(self):
        return self.get_properties()

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.name)

    def __str__(self):
        return self.__repr__()



class Role(object):
    def __init__(self, data):
        self.id = data.get('id')
        self.name = data.get('name')
        self.description = data.get('description')

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.name)

    def __str__(self):
        return self.__repr__()





import xmatters.utils


class ErrorObject(object):
    """ xMatters Error common object"""

    def __init__(self, data):
        self.code = data.get('code')
        self.subcode = data.get('subcode')
        self.reason = data.get('reason')
        self.message = data.get('message')

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.code)

    def __str__(self):
        return self.__repr__()


class Error(Exception):
    pass


class XMSessionError(Error):
    def __init__(self, msg):
        super(XMSessionError, self).__init__(msg)

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class AuthorizationError(Error):
    def __init__(self, msg):
        super(AuthorizationError, self).__init__(msg)


class ApiError(Error):
    def __init__(self, data):
        msg = ' '.join(['{}: {}'.format(k, v) for k, v in data.items()])
        super(ApiError, self).__init__(msg)


class NoContentError(ApiError):
    def __init__(self, data):
        super(NoContentError, self).__init__(data)


class BadRequestError(ApiError):
    def __init__(self, data):
        super(BadRequestError, self).__init__(data)


class UnauthorizedError(ApiError):
    def __init__(self, data):
        super(UnauthorizedError, self).__init__(data)


class ForbiddenError(ApiError):
    def __init__(self, data):
        super(ForbiddenError, self).__init__(data)


class NotFoundError(ApiError):
    def __init__(self, data):
        super(NotFoundError, self).__init__(data)


class NotAcceptableError(ApiError):
    def __init__(self, data):
        super(NotAcceptableError, self).__init__(data)


class ConflictError(ApiError):
    def __init__(self, data):
        super(ConflictError, self).__init__(data)


class UnsupportedMediaError(ApiError):
    def __init__(self, data):
        super(UnsupportedMediaError, self).__init__(data)


class TooManyRequestsError(ApiError):
    def __init__(self, data):
        super(TooManyRequestsError, self).__init__(data)


class ErrorFactory(xmatters.utils.Factory):
    needs_parent = False
    identifier_field = 'code'
    factory_objects = {204: NoContentError,
                       400: BadRequestError,
                       401: UnauthorizedError,
                       403: ForbiddenError,
                       404: NotFoundError,
                       406: NotAcceptableError,
                       409: ConflictError,
                       415: UnsupportedMediaError,
                       429: TooManyRequestsError}

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()



from xmatters.xm_objects.common import Recipient, Pagination, SelfLink
from xmatters.xm_objects.people import Person
from xmatters.xm_objects.roles import Role
from xmatters.xm_objects.subscriptions import SubscriptionCriteriaReference


class DynamicTeam(Recipient):
    _endpoints = {'supervisors': '?embed=supervisors',
                  'observers': '?embed=observers',
                  'get_members': '/members'}

    def __init__(self, parent, data):
        super(DynamicTeam, self).__init__(parent, data)
        self.response_count = data.get('responseCount')
        self.response_count_threshold = data.get('responseCountThreshold')
        self.use_emergency_device = data.get('useEmergencyDevice')
        self.description = data.get('description')
        criteria = data.get('criteria')
        self.criteria = SubscriptionCriteriaReference(criteria) if criteria else None
        links = data.get('links')
        self.links = SelfLink(self, links) if links else None

    @property
    def observers(self):
        url = self.build_url(self._endpoints.get('observers'))
        observers = self.con.get(url).get('observers', {}).get('data')
        return [Role(role) for role in observers] if observers else []

    @property
    def supervisors(self):
        url = self.build_url(self._endpoints.get('supervisors'))
        supervisors = self.con.get(url).get('supervisors', {})
        return Pagination(self, supervisors, Person) if supervisors.get('data') else []

    def get_members(self):
        url = self.build_url(self._endpoints.get('get_members'))
        data = self.con.get(url)
        return Pagination(self, data, Person) if data.get('data') else None

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.target_name)

    def __str__(self):
        return self.__repr__()





import xmatters.connection
import xmatters.utils
import xmatters.factories
import xmatters.xm_objects.forms
from xmatters.xm_objects.common import Pagination, SelfLink
from xmatters.xm_objects.people import Person

class SubscriptionCriteriaReference(object):
    def __init__(self, data):
        self.name = data.get('name')
        self.operator = data.get('operator')
        self.value = data.get('value')
        self.values = data.get('values', [])

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class Subscription(xmatters.connection.ApiBridge):
    _endpoints = {'get_subscribers': '/subscribers'}

    def __init__(self, parent, data):
        super(Subscription, self).__init__(parent, data)
        self.id = data.get('id')
        self.name = data.get('name')
        self.description = data.get('description')
        form = data.get('form')
        self.form = xmatters.xm_objects.forms.FormReference(form) if form else None
        owner = data.get('owner')

        self.owner = xmatters.xm_objects.people.PersonReference(self, owner)
        created = data.get('created')
        self.created = xmatters.utils.TimeAttribute(created) if created else None
        self.notification_delay = data.get('notificationDelay')
        criteria = data.get('criteria', {})
        self.criteria = Pagination(self, criteria, SubscriptionCriteriaReference) if criteria.get(
            'data') else []
        r = data.get('recipients', {})
        self.recipients = Pagination(self, r, xmatters.factories.RecipientFactory) if r.get('data') else []
        tdns = data.get('targetDeviceNames', {})
        self.target_device_names = Pagination(self, tdns, xmatters.factories.DeviceNameFactory) if tdns.get('data') else []
        links = data.get('links')
        self.links = SelfLink(self, links) if links else None

    def get_subscribers(self, offset=None, limit=None):
        params = {'offset': offset,
                  'limit': limit}
        url = self.build_url(self._endpoints.get('get_subscribers'))
        subscribers = self.con.get(url, params=params)
        return Pagination(self, subscribers, xmatters.xm_objects.people.Person, limit=limit) if subscribers.get('data') else []

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()



import xmatters.utils
import xmatters.connection
import xmatters.xm_objects.common


class GroupReference(xmatters.connection.ApiBridge):
    def __init__(self, parent, data):
        super(GroupReference, self).__init__(parent, data)
        self.id = data.get('id')
        self.target_name = data.get('targetName')
        self.recipient_type = data.get('recipientType')
        self.group_type = data.get('groupType')
        links = data.get('links')
        self.links = xmatters.xm_objects.common.SelfLink(self, links) if links else None

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.target_name)

    def __str__(self):
        return self.__repr__()


class End(object):
    def __init__(self, data):
        self.end_by = data.get('endBy')
        date = data.get('date')
        self.date = xmatters.utils.TimeAttribute(date) if date else None
        self.repetitions = data.get('repetitions')

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class Rotation(object):
    def __init__(self, data):
        self.type = data.get('type')
        self.direction = data.get('direction')
        self.interval = data.get('interval')
        self.interval_unit = data.get('intervalUnit')
        next_rotation_time = data.get('nextRotationTime')
        self.next_rotation_time = xmatters.utils.TimeAttribute(next_rotation_time) if next_rotation_time else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class ShiftRecurrence(object):
    def __init__(self, data):
        self.frequency = data.get('frequency')
        self.repeat_every = data.get('repeatEvery')
        self.on_days = data.get('onDays')
        self.on = data.get('on')
        self.months = data.get('months')
        self.data_on_month = data.get('dateOfMonth')
        self.day_of_week_classifier = data.get('dayOfWeekClassifier')
        self.day_of_week = data.get('dayOfWeek')
        end = data.get('end')
        self.end = End(end) if end else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class ShiftMember(xmatters.connection.ApiBridge):
    def __init__(self, parent, data):
        super(ShiftMember, self).__init__(parent, data)
        self.position = data.get('position')
        self.delay = data.get('delay')
        self.escalation_type = data.get('escalationType')
        self.in_rotation = data.get('inRotation')
        recipient = data.get('recipient')
        self.recipient = xmatters.xm_objects.common.Recipient(self, recipient) if recipient else None
        shift = data.get('shift')
        self.shift = xmatters.xm_objects.common.ReferenceByIdAndSelfLink(self, shift) if shift else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class Shift(xmatters.connection.ApiBridge):
    _endpoints = {'get_members': '/members'}

    def __init__(self, parent, data):
        super(Shift, self).__init__(parent, data)
        self.id = data.get('id')
        group = data.get('group')
        self.group = GroupReference(self, group) if group else None
        links = data.get('links')
        self.links = xmatters.xm_objects.common.SelfLink(self, links) if links else None
        self.name = data.get('name')
        start = data.get('start')
        self.start = xmatters.utils.TimeAttribute(start) if start else None
        end = data.get('end')
        self.end = xmatters.utils.TimeAttribute(end) if end else None
        self.timezone = data.get('timezone')
        recurrence = data.get('recurrence')
        self.recurrence = ShiftRecurrence(recurrence) if recurrence else None

    def get_members(self):
        url = self.build_url(self._endpoints.get('get_members'))
        members = self.con.get(url)
        return xmatters.xm_objects.common.Pagination(self, members, ShiftMember) if members.get('data') else []

    # TODO: Test
    def add_member(self, data):
        url = self.build_url(self._endpoints.get('get_members'))
        data = self.con.post(url, data=data)
        return ShiftMember(self, data) if data else None

    @property
    def members(self):
        return self.get_members()

    def __repr__(self):
        return '<Shift {}>'.format(self.name)

    def __str__(self):
        return self.__repr__()



import xmatters.utils
import xmatters.xm_objects.people
import xmatters.factories
import xmatters.connection
import xmatters.xm_objects.oncall
import xmatters.xm_objects.roles
import xmatters.xm_objects.shifts
from xmatters.xm_objects.common import Pagination, SelfLink, RecipientReference, Recipient, ReferenceByIdAndSelfLink


class GroupMembershipShiftReference(xmatters.connection.ApiBridge):
    """ Custom object for shift information embedded in a group membership response """

    def __int__(self, parent, data):
        super(GroupMembershipShiftReference, self).__init__(parent, data)
        self.id = data.get('id')
        group = data.get('group')
        self.group = xmatters.xm_objects.shifts.GroupReference(self, group) if group else None
        self.name = data.get('name')
        links = data.get('links')
        self.links = SelfLink(self, links) if links else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class GroupMembership(xmatters.connection.ApiBridge):
    _endpoints = {'shifts': '/groups/{group_id}/members?embed=shifts'}

    def __init__(self, parent, data):
        super(GroupMembership, self).__init__(parent, data)
        group = data.get('group')
        self.group = xmatters.xm_objects.shifts.GroupReference(self, group) if group else None
        member = data.get('member')
        self.member = RecipientReference(self, member) if member else None
        shifts = data.get('shifts', {})
        self.shifts = Pagination(self, shifts, GroupMembershipShiftReference) if shifts.get('data') else []

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class Group(Recipient):
    _endpoints = {'get_supervisors': '/supervisors',
                  'get_oncall': '{base_url}/on-call?groups={group_id}',
                  'observers': '?embed=observers',
                  'get_shifts': '/shifts',
                  'add_member': '/members',
                  'delete_member': '/members/member_id',
                  'get_members': '/members?embed=shifts',
                  'get_shift_by_id': '/shifts/{shift_id}'}

    def __init__(self, parent, data):
        super(Group, self).__init__(parent, data)
        self.allow_duplicates = data.get('allowDuplicates')
        self.description = data.get('description')
        self.observed_by_all = data.get('observedByAll')
        self.response_count = data.get('responseCount')
        self.response_count_threshold = data.get('responseCount')
        self.use_default_devices = data.get('responseCountThreshold')
        created = data.get('created')
        self.created = xmatters.utils.TimeAttribute(created) if created else None
        self.group_type = data.get('groupType')
        site = data.get('site')
        self.site = ReferenceByIdAndSelfLink(self, site) if site else None
        self.services = data.get('services', [])

    @property
    def observers(self):
        url = self.build_url(self._endpoints.get('observers'))
        observers = self.con.get(url).get('observers', {}).get('data')
        return [xmatters.xm_objects.roles.Role(role) for role in observers] if observers else []

    @property
    def supervisors(self):
        return self.get_supervisors()

    def get_supervisors(self):
        url = self.build_url(self._endpoints.get('get_supervisors'))
        data = self.con.get(url)
        return Pagination(self, data, xmatters.xm_objects.people.Person) if data.get('data') else []

    def get_oncall(self, params=None):
        url = self._endpoints.get('get_oncall').format(base_url=self.con.base_url, group_id=self.id)
        data = self.con.get(url, params=params)
        return Pagination(self, data, xmatters.xm_objects.oncall.OnCall) if data.get('data') else []

    def get_shifts(self, at=None):
        params = {'at': self.process_time_param(at)}
        url = self.build_url(self._endpoints.get('get_shifts'))
        data = self.con.get(url, params)
        return Pagination(self, data, xmatters.xm_objects.shifts.Shift) if data.get('data') else []

    def get_shift_by_id(self, shift_id, at=None):
        params = {'at': self.process_time_param(at)}
        url = self.build_url(self._endpoints.get('get_shift_by_id').format(shift_id=shift_id))
        data = self.con.get(url, params=params)
        return xmatters.xm_objects.shifts.Shift(self, data) if data else None

    # TODO: Test
    def create_shift(self, data):
        url = self.build_url(self._endpoints.get('get_shifts'))
        data = self.con.post(url, data=data)
        return xmatters.xm_objects.shifts.Shift(self, data) if data else None

    # TODO: Test
    def delete_shift(self, shift_id):
        url = self.build_url(self._endpoints.get('get_shift_by_id').format(shift_id=shift_id))
        data = self.con.delete(url)
        return xmatters.xm_objects.shifts.Shift(self, data) if data else None

    def get_members(self):
        url = self.build_url(self._endpoints.get('get_members'))
        data = self.con.get(url)
        return Pagination(self, data, GroupMembership) if data.get('data') else []

    # TODO: Test
    def add_member(self, data):
        url = self.build_url(self._endpoints.get('add_member'))
        data = self.con.post(url, data=data)
        return xmatters.factories.RecipientFactory.compose(self, data) if data else None

    # TODO: Test
    def remove_member(self, member_id):
        url = self.build_url(self._endpoints.get('remove_member').format(member_id=member_id))
        data = self.con.delete(url)
        return GroupMembership(self, data) if data else None

    def get_observers(self):
        return self.observers

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.target_name)

    def __str__(self):
        return self.__repr__()



from xmatters.xm_objects.common import SelfLink
from xmatters.connection import ApiBridge
import xmatters.xm_objects.events as events
import xmatters.utils as util


class EventFloodFilter(object):
    def __init__(self, data):
        self.id = data.get('id')
        self.name = data.get('name')

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class SuppressionMatch(ApiBridge):
    def __init__(self, parent, data):
        super(SuppressionMatch, self).__init__(parent, data)
        self.id = data.get('id')
        self.event_id = data.get('eventId')
        links = data.get('links')
        self.links = SelfLink(self, links) if links else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class EventSuppression(ApiBridge):
    def __init__(self, parent, data):
        super(EventSuppression, self).__init__(parent, data)
        event = data.get('event')
        self.event = events.EventReference(self, event) if event else None
        match = data.get('match')
        self.match = SuppressionMatch(self, data) if match else None
        at = data.get('at')
        self.at = util.TimeAttribute(at) if at else None
        filters = data.get('filters', [])
        self.filter = [EventFloodFilter(f) for f in filters]
        self.links = SelfLink(self, data.get('links'))

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()





from xmatters.xm_objects.common import ReferenceById, Pagination, SelfLink
from xmatters.xm_objects.people import PersonReference
import xmatters.xm_objects.plan_endpoints
import xmatters.xm_objects.plans
from xmatters.connection import ApiBridge


class IntegrationReference(object):
    def __init__(self, data):
        self.id = data.get('id')
        plan = data.get('plan')
        self.plan = xmatters.xm_objects.plans.PlanReference(plan) if plan else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class IntegrationLog(ApiBridge):
    def __init__(self, parent, data):
        super(IntegrationLog, self).__init__(parent, data)
        self.id = data.get('id')
        integration = data.get('integration')
        self.integration = IntegrationReference(integration) if integration else None
        self.completed = data.get('completed')
        self.request_method = data.get('requestMethod')
        self.request_headers = data.get('requestHeaders')
        self.request_parameters = data.get('requestParameters')
        self.request_body = data.get('requestBody')
        self.remote_address = data.get('remoteAddress')
        self.request_id = data.get('requestId')
        self.status = data.get('status')
        by = data.get('by')
        self.by = PersonReference(self, by) if by else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class Integration(ApiBridge):
    _endpoints = {'get_logs': '/logs'}

    def __init__(self, parent, data):
        super(Integration, self).__init__(parent, data)
        self.id = data.get('id')
        plan = data.get('plan')
        self.plan = ReferenceById(plan) if plan else None
        form = data.get('form')
        self.form = ReferenceById(form) if form else None
        self.name = data.get('name')
        self.integration_type = data.get('integrationType')
        self.operation = data.get('operation')
        self.triggered_by = data.get('triggeredBy')
        self.created_by = data.get('createdBy')
        self.authentication_type = data.get('authenticationType')
        endpoint = data.get('endpoint')
        self.endpoint = xmatters.xm_objects.plan_endpoints.Endpoint(self, endpoint) if endpoint else None
        self.deployed = data.get('deployed')
        self.script = data.get('script')
        self.migrated_outbound_trigger = data.get('migratedOutboundTrigger')
        self.origin_type = data.get('originType')
        self.is_run_by_service_owner = data.get('isRunByServiceOwner')
        links = data.get('links')
        self.links = SelfLink(self, links) if links else None

    # TODO: Test
    def get_logs(self):
        endpoint = self.build_url(self._endpoints.get('get_logs'))
        url = self.build_url(endpoint)
        logs = self.con.get(url)
        return Pagination(self, logs, IntegrationLog) if logs.get('data') else []

    @property
    def logs(self):
        return self.get_logs()

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.name)

    def __str__(self):
        return self.__repr__()



import urllib.parse

import xmatters.auth
import xmatters.connection
import xmatters.endpoints


class XMSession(object):
    """ Primary class used to interact with xMatters API """

    def __init__(self, base_url, **kwargs):
        """
        :param base_url: xMatters instance url or xMatters instance base url
        :type base_url: str
        """
        p_url = urllib.parse.urlparse(base_url)
        instance_url = 'https://{}'.format(p_url.netloc) if p_url.netloc else 'https://{}'.format(p_url.path)
        self._base_url = '{}/api/xm/1'.format(instance_url)
        self.con = None
        self._kwargs = kwargs

    def set_authentication(self, username=None, password=None, client_id=None, token=None, token_storage=None):
        """
        Set authentication method to use.
            If client_id is provided, it is assumed OAuth2 authentication is desired;
            otherwise basic authentication is used.
        :param username: xMatters username
        :type username: str, optional
        :param password: xMatters password
        :type password: str, optional
        :param client_id: xMatters instance client id
        :type client_id: str, optional
        :param token: xMatters token object or refresh token
        :type token: dict or str, optional
        :param token_storage: Class instance used to store token returned during a refresh.
            Any class instance will be accepted as long as it has "read_token" and "write_token" methods.
        :type token_storage: :class:`xmatters.utils.TokenFileStorage`, optional
        :return: self
        :rtype: :class:`xmatters.session.XMSession`
        """
        timeout = self._kwargs.get('timeout')
        max_retries = self._kwargs.get('max_retries')
        if client_id:
            self.con = xmatters.auth.OAuth2Auth(self._base_url, client_id, token, username, password, token_storage,
                                                timeout=timeout,
                                                max_retries=max_retries)
        elif None not in (username, password):
            self.con = xmatters.auth.BasicAuth(self._base_url, username, password, timeout=timeout,
                                               max_retries=max_retries)
        else:
            raise ValueError('unable to determine authentication method')

        return self

    def audits(self):
        return xmatters.endpoints.AuditsEndpoint(self)

    def conference_bridges(self):
        return xmatters.endpoints.ConferenceBridgesEndpoint(self)

    def device_names(self):
        return xmatters.endpoints.DeviceNamesEndpoint(self)

    def device_types(self):
        return xmatters.endpoints.DeviceTypesEndpoint(self)

    def devices(self):
        return xmatters.endpoints.DevicesEndpoint(self)

    def dynamic_teams(self):
        return xmatters.endpoints.DynamicTeamsEndpoint(self)

    def events(self):
        return xmatters.endpoints.EventsEndpoint(self)

    def event_suppressions(self):
        return xmatters.endpoints.EventSuppressionsEndpoint(self)

    def forms(self):
        return xmatters.endpoints.FormsEndpoint(self)

    def import_jobs(self):
        return xmatters.endpoints.ImportJobsEndpoint(self)

    def groups(self):
        return xmatters.endpoints.GroupsEndpoint(self)

    def incidents(self):
        return xmatters.endpoints.IncidentsEndpoint(self)

    def oncall(self):
        return xmatters.endpoints.OnCallEndpoint(self)

    def oncall_summary(self):
        return xmatters.endpoints.OnCallSummaryEndpoint(self)

    def people(self):
        return xmatters.endpoints.PeopleEndpoint(self)

    def plans(self):
        return xmatters.endpoints.PlansEndpoint(self)

    def roles(self):
        return xmatters.endpoints.RolesEndpoint(self)

    def scenarios(self):
        return xmatters.endpoints.ScenariosEndpoint(self)

    def services(self):
        return xmatters.endpoints.ServicesEndpoint(self)

    def sites(self):
        return xmatters.endpoints.SitesEndpoint(self)

    def subscriptions(self):
        return xmatters.endpoints.SubscriptionsEndpoint(self)

    def subscription_forms(self):
        return xmatters.endpoints.SubscriptionFormsEndpoint(self)

    def temporary_absences(self):
        return xmatters.endpoints.TemporaryAbsencesEndpoint(self)

    # TODO
    # def upload_users(self):
    #     return xmatters.endpoints.UploadUsersEndpoint(self)

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()



import requests
from oauthlib.oauth2 import LegacyApplicationClient
from requests_oauthlib import OAuth2Session

from xmatters import errors as err
from xmatters.connection import Connection


class OAuth2Auth(Connection):
    _endpoints = {'token': '/oauth2/token'}

    def __init__(self, base_url, client_id, token=None, username=None, password=None, token_storage=None, **kwargs):
        """
        Class used to authentication requests using OAuth2 authentication.

        The method used to obtain a token are used in the following order:  token, username & password, token_storage.

        :param base_url: xMatters instance url or xMatters instance base url
        :type base_url: str
        :param client_id: xMatters instance client id
        :type client_id: str
        :param token: Authentication token. Can be just a refresh token (as a str) or a token object (as a dict)
        :type token: str or dict, optional
         :param username: xMatters username
        :type username: str, optional
        :param password: xMatters password
        :type password: str, optional
        :param token_storage: Class instance used to store token returned during a refresh.
            Any class instance will be accepted as long as it has "read_token" and "write_token" methods.
        :type token_storage: :class:`xmatters.utils.TokenFileStorage`, optional
        """
        self.client_id = client_id
        self.token_storage = token_storage
        self.username = username
        self.password = password
        self._token = token
        self.session = None
        token_url = '{}{}'.format(base_url, self._endpoints.get('token'))
        client = LegacyApplicationClient(client_id=self.client_id)
        auto_refresh_kwargs = {'client_id': self.client_id}
        token_updater = self.token_storage.write_token if self.token_storage else None
        self.session = OAuth2Session(client=client, auto_refresh_url=token_url,
                                     auto_refresh_kwargs=auto_refresh_kwargs,
                                     token_updater=token_updater)
        self.session.token = self.get_token()

        # update storage token if differs from self.token
        if self.token_storage and self.token_storage.read_token() != self.token:
            self.token_storage.write_token(self.token)

        super(OAuth2Auth, self).__init__(base_url, self.session, **kwargs)

    def get_token(self):
        if self._token and isinstance(self._token, dict):
            return self._token
        elif self._token and isinstance(self._token, str):
            return self.session.refresh_token(token_url=self.session.auto_refresh_url, refresh_token=self._token,
                                              timeout=3,
                                              kwargs=self.session.auto_refresh_kwargs)
        elif None not in (self.username, self.password):
            return self.session.fetch_token(token_url=self.session.auto_refresh_url, username=self.username,
                                            password=self.password, include_client_id=True, timeout=3)
        elif self.token_storage and self.token_storage.read_token():
            return self.token_storage.read_token()
        else:
            raise err.XMSessionError('Unable to obtain token with provided arguments')

    @property
    def token(self):
        return self.session.token if self.session else self._token

    @token.setter
    def token(self, token):
        if self.session:
            self.session.token = token
        else:
            self._token = token

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class BasicAuth(Connection):
    def __init__(self, base_url, username, password, **kwargs):
        """
        Class used to authentication requests using basic authentication

        :param base_url: xMatters instance url or xMatters instance base url
        :type base_url: str
        :param username: xMatters username
        :type username: str
        :param password: xMatters password
        :type password: str
        """
        self.username = username
        self.password = password
        self.session = requests.Session()
        self.session.auth = (self.username, self.password)
        super(BasicAuth, self).__init__(base_url, self.session, **kwargs)

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()



import xmatters.xm_objects.groups
import xmatters.xm_objects.audits
import xmatters.xm_objects.people
import xmatters.xm_objects.dynamic_teams
import xmatters.xm_objects.scenarios
import xmatters.xm_objects.device_names
import xmatters.xm_objects.plan_endpoints
import xmatters.xm_objects.plan_properties
import xmatters.xm_objects.forms
import xmatters.xm_objects.devices
import xmatters.utils


class DeviceFactory(xmatters.utils.Factory):
    needs_parent = True
    identifier_field = 'deviceType'
    factory_objects = {'EMAIL': xmatters.xm_objects.devices.EmailDevice,
                       'VOICE': xmatters.xm_objects.devices.VoiceDevice,
                       'TEXT_PHONE': xmatters.xm_objects.devices.SMSDevice,
                       'TEXT_PAGER': xmatters.xm_objects.devices.TextPagerDevice,
                       'APPLE_PUSH': xmatters.xm_objects.devices.ApplePushDevice,
                       'ANDROID_PUSH': xmatters.xm_objects.devices.AndroidPushDevice,
                       'FAX': xmatters.xm_objects.devices.FaxDevice,
                       'VOICE_IVR': xmatters.xm_objects.devices.PublicAddressDevice,
                       'GENERIC': xmatters.xm_objects.devices.GenericDevice}

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class RecipientFactory(xmatters.utils.Factory):
    needs_parent = True
    identifier_field = 'recipientType'
    factory_objects = {'GROUP': xmatters.xm_objects.groups.Group,
                       'PERSON': xmatters.xm_objects.people.Person,
                       'DEVICE': xmatters.xm_objects.devices.Device,
                       'DYNAMIC_TEAM': xmatters.xm_objects.dynamic_teams.DynamicTeam}

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class AuditFactory(xmatters.utils.Factory):
    needs_parent = True
    identifier_field = 'type'
    factory_objects = {'EVENT_ANNOTATED': xmatters.xm_objects.audits.AuditAnnotation,
                       'EVENT_CREATED': xmatters.xm_objects.audits.Audit,
                       'EVENT_SUSPENDED': xmatters.xm_objects.audits.Audit,
                       'EVENT_RESUMED': xmatters.xm_objects.audits.Audit,
                       'EVENT_COMPLETED': xmatters.xm_objects.audits.Audit,
                       'EVENT_TERMINATED': xmatters.xm_objects.audits.Audit,
                       'RESPONSE_RECEIVED': xmatters.xm_objects.audits.AuditResponse,
                       'NOTIFICATION_DELIVERED': xmatters.xm_objects.audits.AuditNotification,
                       'NOTIFICATION_FAILED': xmatters.xm_objects.audits.AuditNotification}

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class SectionFactory(xmatters.utils.Factory):
    needs_parent = True
    identifier_field = 'type'
    factory_objects = {'CONFERENCE_BRIDGE': xmatters.xm_objects.forms.ConferenceBridgeSection,
                       'CUSTOM_SECTION': xmatters.xm_objects.forms.CustomSectionItems,
                       'DEVICE_FILTER': xmatters.xm_objects.forms.DevicesSection,
                       'HANDLING_OPTIONS': xmatters.xm_objects.forms.HandlingSection,
                       'ATTACHMENTS': xmatters.xm_objects.forms.FormSection,
                       'SENDER_OVERRIDES': xmatters.xm_objects.forms.SenderOverridesSection,
                       'RECIPIENTS': xmatters.xm_objects.forms.RecipientsSection,
                       'RESPONSE_CHOICES': xmatters.xm_objects.forms.FormSection,
                       'INCIDENT': xmatters.xm_objects.forms.IncidentSection,
                       'DOCUMENT_UPLOAD': xmatters.xm_objects.forms.FormSection}

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class AuthFactory(xmatters.utils.Factory):
    needs_parent = False
    identifier_field = 'authenticationType'
    factory_objects = {'NO_AUTH': None,
                       'BASIC': xmatters.xm_objects.plan_endpoints.BasicAuthentication,
                       'OAUTH2': xmatters.xm_objects.plan_endpoints.OAuth2Authentication,
                       'OAUTH2_FORCE': xmatters.xm_objects.plan_endpoints.OAuth2Authentication,
                       'OAUTH_SLACK': xmatters.xm_objects.plan_endpoints.OAuth2Authentication,
                       'XMATTERS': xmatters.xm_objects.plan_endpoints.ServiceAuthentication,
                       'SERVICENOW': xmatters.xm_objects.plan_endpoints.ServiceAuthentication}

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class PropertiesFactory(xmatters.utils.Factory):
    needs_parent = False
    identifier_field = 'propertyType'
    factory_objects = {'BOOLEAN': xmatters.xm_objects.plan_properties.Boolean,
                       'HIERARCHY': xmatters.xm_objects.plan_properties.Hierarchy,
                       'LIST_TEXT_MULTI_SELECT': xmatters.xm_objects.plan_properties.MultLinkSelectList,
                       'LIST_TEXT_SINGLE_SELECT': xmatters.xm_objects.plan_properties.SingleSelectList,
                       'NUMBER': xmatters.xm_objects.plan_properties.Number,
                       'PASSWORD': xmatters.xm_objects.plan_properties.Password,
                       'TEXT': xmatters.xm_objects.plan_properties.Text}

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class ScenarioPermFactory(xmatters.utils.Factory):
    needs_parent = True
    identifier_field = 'permissibleType'
    factory_objects = {'PERSON': xmatters.xm_objects.scenarios.ScenarioPermissionPerson,
                       'ROLE': xmatters.xm_objects.scenarios.ScenarioPermissionRole}

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class DeviceNameFactory(xmatters.utils.Factory):
    needs_parent = False
    identifier_field = 'deviceType'
    factory_objects = {
        'ANDROID_PUSH': xmatters.xm_objects.device_names.DeviceName,
        'APPLE_PUSH': xmatters.xm_objects.device_names.DeviceName,
        'EMAIL': xmatters.xm_objects.device_names.DeviceNameEmail,
        'FAX': xmatters.xm_objects.device_names.DeviceName,
        'GENERIC': xmatters.xm_objects.device_names.DeviceName,
        'TEXT_PAGER': xmatters.xm_objects.device_names.DeviceName,
        'TEXT_PHONE': xmatters.xm_objects.device_names.DeviceName,
        'VOICE': xmatters.xm_objects.device_names.DeviceName,
        'VOICE_IVR': xmatters.xm_objects.device_names.DeviceName}

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()



import json
import pathlib
from abc import ABC, abstractmethod

from dateutil import tz, parser

MAX_API_LIMIT = 1000


class TimeAttribute(str):
    def datetime(self):
        """
        Create datetime object from str
        :return: datetime object with whichever timezone in str on creation (if present)
        :rtype: :class:`datetime.datetime`
        """
        return parser.isoparse(self)

    def isoformat(self):
        return self.datetime().isoformat()

    def datetime_local(self):
        """
        Create datetime object from str and adjust to local timezone
        :return: datetime object adjusted to local timezone
        :rtype: :class:`datetime.datetime`
        """
        return parser.isoparse(self).astimezone(tz.tzlocal())

    def isoformat_local(self):
        return self.datetime_local().isoformat()

    def datetime_utc(self):
        """
        Create datetime object from str and adjust to utc timezone.
        :return: datetime object adjusted to utc timezone
        :rtype: :class:`datetime.datetime`
        """
        return parser.isoparse(self).astimezone(tz.tzutc())

    def isoformat_utc(self):
        return self.datetime_utc().isoformat()

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self)


class TokenFileStorage(object):
    def __init__(self, token_filepath):
        if not isinstance(token_filepath, pathlib.Path):
            self.token_filepath = pathlib.Path(token_filepath)

    def read_token(self):
        if not self.token_filepath.is_file():
            return None
        else:
            with open(self.token_filepath, 'r') as f:
                return json.load(f)

    def write_token(self, token):
        with open(self.token_filepath, 'w') as f:
            json.dump(token, f, indent=4)

    @property
    def token(self):
        return self.read_token()

    @token.setter
    def token(self, token):
        self.write_token(token)

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class Factory(ABC):

    @property
    @abstractmethod
    def needs_parent(self):
        pass

    @property
    @abstractmethod
    def identifier_field(self):
        pass

    @property
    @abstractmethod
    def factory_objects(self):
        pass

    @classmethod
    def compose(cls, parent, item_data, default=None):
        identifier = item_data.get(cls.identifier_field)
        # noinspection PyUnresolvedReferences
        constructor = cls.factory_objects.get(identifier, default)

        if cls.needs_parent:
            return constructor(parent, item_data) if constructor else None
        else:
            return constructor(item_data) if constructor else None



import xmatters.utils as util
import xmatters.factories
from xmatters.xm_objects.common import SelfLink, Pagination
from xmatters.connection import ApiBridge
from xmatters.xm_objects.shifts import GroupReference, Shift


class Replacer(ApiBridge):
    def __init__(self, parent, data):
        super(Replacer, self).__init__(parent, data)
        self.id = data.get('id')
        self.target_name = data.get('targetName')
        self.recipient_type = data.get('recipientType')
        links = data.get('links')
        self.links = SelfLink(self, links) if links else None
        self.first_name = data.get('firstName')
        self.last_name = data.get('lastName')
        self.status = data.get('status')

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.target_name)

    def __str__(self):
        return self.__repr__()


class ShiftOccurrenceMember(ApiBridge):
    def __init__(self, parent, data):
        super(ShiftOccurrenceMember, self).__init__(parent, data)
        member = data.get('member')
        self.member = xmatters.factories.RecipientFactory.compose(self, member) if member else None
        self.position = data.get('position')
        self.delay = data.get('delay')
        self.escalation_type = data.get('escalationType')
        replacements = data.get('replacements', {})
        self.replacements = Pagination(self, replacements, TemporaryReplacement) if replacements.get('data') else []

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.member.target_name)

    def __str__(self):
        return self.__repr__()


class ShiftReference(ApiBridge):
    def __init__(self, parent, data):
        super(ShiftReference, self).__init__(parent, data)
        self.id = data.get('id')
        links = data.get('links')
        self.links = SelfLink(self, links) if links else None
        self.name = data.get('name')

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.name)

    def __str__(self):
        return self.__repr__()


class TemporaryReplacement(ApiBridge):
    def __init__(self, parent, data):
        super(TemporaryReplacement, self).__init__(parent, data)
        start = data.get('start')
        self.start = util.TimeAttribute(start) if start else None
        end = data.get('end')
        self.end = util.TimeAttribute(end) if end else None
        replacement = data.get('replacement')
        self.replacement = TemporaryReplacement(self, replacement) if replacement else None


class OnCall(ApiBridge):
    def __init__(self, parent, data):
        super(OnCall, self).__init__(parent)
        # save shift self link for use with 'shift' property to return full Shift object (not just ShiftReference)
        self._shift_link = data.get('shift', {}).get('links', {}).get('self')
        group = data.get('group')
        self.group = GroupReference(parent, group) if group else None
        start = data.get('start')
        self.start = util.TimeAttribute(start) if start else None
        end = data.get('end')
        self.end = util.TimeAttribute(end) if end else None
        members = data.get('members', {})
        self.members = Pagination(self, members, ShiftOccurrenceMember) if members.get('data') else []

    @property
    def shift(self):
        if self._shift_link:
            url = '{}{}'.format(self.con.instance_url, self._shift_link)
            data = self.con.get(url)
            return Shift(self, data) if data else None
        else:
            return None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()



import xmatters.connection
import xmatters.xm_objects.shifts


class Service(xmatters.connection.ApiBridge):
    def __init__(self, parent, data):
        super(Service, self).__init__(parent, data)
        self.id = data.get('id')
        self.target_name = data.get('targetName')
        self.recipients_type = data.get('recipientType')
        self.description = data.get('description')
        owned_by = data.get('ownedBy')
        self.owned_by = xmatters.xm_objects.shifts.GroupReference(self, owned_by) if owned_by else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()



import xmatters.xm_objects.events
import xmatters.xm_objects.forms
import xmatters.factories as factory
import xmatters.utils
import xmatters.connection
import xmatters.xm_objects.people
import xmatters.xm_objects.plans
import xmatters.xm_objects.roles

from xmatters.xm_objects.common import Pagination, SelfLink


class ScenarioPermission(xmatters.connection.ApiBridge):
    def __init__(self, parent, data):
        super(ScenarioPermission, self).__init__(parent, data)
        self.permissible_type = data.get('permissibleType')
        self.editor = data.get('editor')

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class ScenarioPermissionPerson(ScenarioPermission):
    def __init__(self, parent, data):
        super(ScenarioPermissionPerson, self).__init__(parent, data)
        person = data.get('person')
        self.person = xmatters.xm_objects.people.PersonReference(self, person) if person else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class ScenarioPermissionRole(ScenarioPermission):
    def __init__(self, parent, data):
        super(ScenarioPermissionRole, self).__init__(parent, data)
        role = data.get('role')
        self.role = xmatters.xm_objects.roles.Role(role) if role else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class Scenario(xmatters.connection.ApiBridge):
    _endpoints = {'properties': '?embed=properties',
                  'plan': '?embed=properties',
                  'form': '?embed=form',
                  'properties_translations': '?embed=properties.translations'}

    def __init__(self, parent, data):
        super(Scenario, self).__init__(parent, data)
        self.id = data.get('id')
        self.name = data.get('name')
        self.description = data.get('description')
        self.priority = data.get('priority')
        self.position = data.get('position')
        self.bypass_phone_intro = data.get('bypassPhoneIntro')
        self.escalation_override = data.get('escalationOverride')
        self.expiration_in_minutes = data.get('expirationInMinutes')
        self.override_device_restrictions = data.get('overrideDeviceRestrictions')
        self.require_phone_password = data.get('requirePhonePassword')
        sos = data.get('senderOverrides')
        self.sender_overrides = xmatters.xm_objects.forms.SenderOverrides(sos) if sos else None
        vm_opts = data.get('voicemailOptions')
        self.voicemail_options = xmatters.xm_objects.events.VoicemailOptions(vm_opts) if vm_opts else None
        tdns = data.get('targetDeviceNames', {})
        self.target_device_names = Pagination(self, tdns, factory.DeviceNameFactory) if tdns.get('data') else []
        created = data.get('created')
        self.created = xmatters.utils.TimeAttribute(created) if created else None
        perm = data.get('permitted', {}).get('data')
        self.permitted = [factory.ScenarioPermFactory.compose(self, p) for p in perm] if perm else []
        rs = data.get('recipients')
        self.recipients = Pagination(self, rs, factory.RecipientFactory) if rs.get('data') else []
        links = data.get('links')
        self.links = SelfLink(self, links) if links else None

    @property
    def properties(self):
        url = self.build_url(self._endpoints.get('properties'))
        data = self.con.get(url)
        return data.get('properties', {})

    @property
    def plan(self):
        url = self.build_url(self._endpoints.get('plan'))
        plan = self.con.get(url).get('plan', {})
        return xmatters.xm_objects.plans.Plan(self, plan) if plan else None

    @property
    def form(self):
        url = self.build_url(self._endpoints.get('form'))
        form = self.con.get(url).get('form', {})
        return xmatters.xm_objects.forms.Form(self, form) if form else None

    @property
    def properties_translations(self):
        url = self.build_url(self._endpoints.get('properties_translations'))
        data = self.con.get(url)
        return data.get('properties', {})

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()



import xmatters.factories
import xmatters.xm_objects.forms
import xmatters.utils
import xmatters.connection
import xmatters.xm_objects.plans
import xmatters.xm_objects.roles
from xmatters.xm_objects.common import Pagination, SelfLink

class SubscriptionForm(xmatters.connection.ApiBridge):
    _endpoints = {'target_device_names': '?embed=deviceNames',
                  'visible_target_device_names': '?embed=deviceNames',
                  'property_definitions': '?embed=propertyDefinitions',
                  'roles': '?embed=roles'}

    def __init__(self, parent, data):
        super(SubscriptionForm, self).__init__(parent, data)
        self.id = data.get('id')
        self.name = data.get('name')
        self.description = data.get('description')
        plan = data.get('plan')
        self.plan = xmatters.xm_objects.plans.PlanReference(data) if plan else None
        self.scope = data.get('scope')
        form = data.get('form')
        self.form = xmatters.xm_objects.forms.FormReference(form) if form else None
        created = data.get('created')
        self.created = xmatters.utils.TimeAttribute(created) if created else None
        self.one_way = data.get('oneWay')
        self.subscribe_others = data.get('subscribeOthers')
        self.notification_delay = data.get('notificationDelay')
        links = data.get('links')
        self.links = SelfLink(self, links) if links else None

    @property
    def target_device_names(self):
        url = self.build_url(self._endpoints.get('target_device_names'))
        data = self.con.get(url)
        tdns = data.get('targetDeviceNames', {})
        return Pagination(self, tdns, xmatters.factories.DeviceNameFactory) if tdns.get('data') else []

    @property
    def visible_target_device_names(self):
        url = self.build_url(self._endpoints.get('visible_target_device_names'))
        data = self.con.get(url)
        vtdns = data.get('visibleTargetDeviceNames', {})
        return Pagination(self, vtdns, xmatters.factories.DeviceNameFactory) if vtdns.get('data') else []

    @property
    def property_definitions(self):
        url = self.build_url(self._endpoints.get('property_definitions'))
        data = self.con.get(url)
        ps = data.get('propertyDefinitions', {})
        return Pagination(self, ps, xmatters.factories.PropertiesFactory) if ps.get('data') else []

    @property
    def roles(self):
        url = self.build_url(self._endpoints.get('roles'))
        data = self.con.get(url).get('roles')
        roles = data.get('roles')
        return Pagination(self, roles, xmatters.xm_objects.roles.Role) if roles else []

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.name)

    def __str__(self):
        return self.__repr__()


class SubscriptionFormReference(object):
    def __init__(self, data):
        self.id = data.get('id')
        self.name = data.get('name')
        plan = data.get('plan')
        self.plan = xmatters.xm_objects.plans.PlanReference(plan) if plan else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()



import xmatters.utils
import xmatters.xm_objects.people
import xmatters.xm_objects.shifts
import xmatters.connection


class TemporaryAbsence(xmatters.connection.ApiBridge):
    def __init__(self, parent, data):
        super(TemporaryAbsence, self).__init__(parent, data)
        self.id = data.get('id')
        self.absence_type = data.get('absenceType')
        member = data.get('member')
        self.member = xmatters.xm_objects.people.PersonReference(self, member) if member else None
        start = data.get('start')
        self.start = xmatters.utils.TimeAttribute(start) if start else None
        end = data.get('end')
        self.end = xmatters.utils.TimeAttribute(end) if end else None
        group = data.get('group')
        self.group = xmatters.xm_objects.shifts.GroupReference(self, group) if group else None
        replacement = data.get('replacement')
        self.replacement = xmatters.xm_objects.people.PersonReference(self, replacement) if replacement else None

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.member.target_name)

    def __str__(self):
        return self.__repr__()





class Property(object):
    def __init__(self, data):
        self.id = data.get('id')
        self.property_type = data.get('propertyType')
        self.name = data.get('name')
        self.description = data.get('description')
        self.help_text = data.get('helpText')

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class Boolean(Property):
    def __init__(self, data):
        super(Boolean, self).__init__(data)
        self.default = data.get('default')

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class Hierarchy(Property):
    def __init__(self, data):
        super(Hierarchy, self).__init__(data)
        self.default = data.get('default')
        self.delimiter = data.get('delimiter')
        self.categories = data.get('categories', [])
        self.paths = data.get('paths', [])

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class MultLinkSelectList(Property):
    def __init__(self, data):
        super(MultLinkSelectList, self).__init__(data)
        self.items = data.get('items', [])

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class SingleSelectList(Property):
    def __init__(self, data):
        super(SingleSelectList, self).__init__(data)
        self.items = data.get('items', [])

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class Number(Property):
    def __init__(self, data):
        super(Number, self).__init__(data)
        self.default = data.get('default')
        self.max_length = data.get('maxLength')
        self.min_length = data.get('minLength')
        self.units = data.get('units')

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class Password(Property):
    def __init__(self, data):
        super(Password, self).__init__(data)
        self.max_length = data.get('maxLength')
        self.min_length = data.get('minLength')
        self.pattern = data.get('pattern')
        self.validate = data.get('validate')

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class Text(Property):
    def __init__(self, data):
        super(Text, self).__init__(data)
        self.max_length = data.get('maxLength')
        self.min_length = data.get('minLength')
        self.pattern = data.get('pattern')
        self.validate = data.get('validate')

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()



import xmatters.utils
import xmatters.connection
import xmatters.xm_objects.people
import xmatters.xm_objects.common


class ImportMessage(object):
    def __init__(self, data):
        self.id = data.get('id')
        self.message_level = data.get('messageLevel')
        self.message_type = data.get('messageType')
        self.description = data.get('description')
        self.line = data.get('line')

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.message_type)

    def __str__(self):
        return self.__repr__()


class Import(xmatters.connection.ApiBridge):
    _endpoints = {'get_messages': '/import-messages'}

    def __init__(self, parent, data):
        super(Import, self).__init__(parent, data)
        self.id = data.get('id')
        self.name = data.get('name')
        self.transform = data.get('transform')
        self.status = data.get('status')
        started = data.get('started')
        self.started = xmatters.utils.TimeAttribute(started) if started else None
        self.last_updated_at = data.get('lastUpdatedAt')
        by = data.get('by')
        self.by = xmatters.xm_objects.people.PersonReference(parent, by) if by else None
        self.total_count = data.get('totalCount')
        self.processed_count = data.get('processedCount')
        finished_at = data.get('finishedAt')
        self.finished_at = xmatters.utils.TimeAttribute(finished_at) if finished_at else None
        links = data.get('links')
        self.links = xmatters.xm_objects.common.SelfLink(parent, links)

    def get_messages(self, params=None):
        url = self.build_url(self._endpoints.get('get_messages'))
        messages = self.con.get(url, params).get('data', None)
        return [ImportMessage(m) for m in messages] if messages else []

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.name)

    def __str__(self):
        return self.__repr__()



class DeviceTypes(object):
    def __init__(self, data):
        self.count = data.get('count')
        self.total = data.get('total')
        self.data = data.get('data', [])

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()





import xmatters.xm_objects.plans


class SharedLibrary(object):
    def __init__(self, data):
        self.id = data.get('id')
        self.name = data.get('name')
        self.script = data.get('script')
        plan = data.get('plan')
        self.plan = xmatters.xm_objects.plans.PlanReference(plan) if plan else None



import os

from setuptools import setup, find_packages

VERSION = '1.0.3'

# Available classifiers: https://pypi.org/pypi?%3Aaction=list_classifiers
CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Topic :: System :: Monitoring',
    'Topic :: Software Development :: Libraries',
    'Topic :: Communications',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3 :: Only',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Operating System :: OS Independent',
]


def read(filename):
    with open(os.path.join(os.path.dirname(__file__), filename), 'r') as file:
        return file.read()


requires = [
    'certifi>=2021.10.8',
    'charset-normalizer>=2.0.11',
    'idna>=3.3',
    'oauthlib>=3.2.0',
    'python-dateutil>=2.8.2',
    'requests>=2.27.1',
    'requests-oauthlib>=1.3.1',
    'six>=1.16.0',
    'urllib3>=1.26.8'
]

setup(
    name='python-xmatters',
    version=VERSION,
    packages=find_packages(exclude=("tests",)),
    url='https://github.com/hcallen/python-xmatters',
    license='MIT',
    author='hcallen',
    author_email='hallieallen@gmail.com',
    maintainer='hcallen',
    maintainer_email='hallieallen@gmail.com',
    description='Python xMatters API wrapper',
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    classifiers=CLASSIFIERS,
    python_requires=">=3.5",
    install_requires=requires,
    setup_requires=["wheel"],
)



import xmatters.utils as utils
from xmatters.xm_objects.common import SelfLink
from xmatters.xm_objects.people import PersonReference
from xmatters.connection import ApiBridge


class IncidentProperty(object):

    def __init__(self, data):
        self.name = data.get('name')
        self.level = data.get('level')

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.name)

    def __str__(self):
        return self.__repr__()


class Incident(ApiBridge):
    _endpoints = {'add_timeline_note': '/timeline-entries'}

    def __init__(self, parent, data):
        super(Incident, self).__init__(parent, data)
        self.id = data.get('id')
        self.incident_identifier = data.get('incidentIdentifier')
        self.summary = data.get('summary')
        self.description = data.get('description')
        severity = data.get('severity')
        self.severity = IncidentProperty(severity) if severity else None
        status = data.get('status')
        self.status = IncidentProperty(status) if status else None
        initiated_by = data.get('initiatedBy')
        self.initiated_by = PersonReference(self, initiated_by) if initiated_by else None
        commander = data.get('commander')
        self.commander = PersonReference(self, commander) if commander else None
        self.request_id = data.get('requestId')
        self.impacted_services = data.get('impactedServices')
        reporter = data.get('reporter')
        self.reporter = PersonReference(self, reporter) if reporter else None
        created_at = data.get('createdAt')
        self.created_at = utils.TimeAttribute(created_at) if created_at else None
        updated_at = data.get('updated_at')
        self.updated_at = utils.TimeAttribute(updated_at) if updated_at else None
        acknowledged_at = data.get('acknowledgeAt')
        self.acknowledged_at = utils.TimeAttribute(acknowledged_at) if acknowledged_at else None
        links = data.get('links')
        self.links = SelfLink(self, links) if links else None

    # TODO: Test
    def add_timeline_note(self, text):
        data = {'entryType': 'TIMELINE_NOTE',
                'text': text}
        url = self.build_url(self._endpoints.get('add_timeline_note'))
        data = self.con.post(url, data=data)
        return IncidentNote(self, data) if data else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class IncidentNote(ApiBridge):
    def __init__(self, parent, data):
        super(IncidentNote, self).__init__(parent, data)
        self.id = data.get('id')
        at = data.get('at')
        self.at = utils.TimeAttribute(at) if at else None
        self.entry_type = data.get('entryType')
        self.text = data.get('text')
        added_by = data.get('addedBy')
        self.added_dy = PersonReference(self, added_by) if added_by else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class IncidentDetails(object):
    def __init__(self, data):
        self.summary = data.get('summary')
        self.description = data.get('description')
        self.severity = data.get('severity')

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()



import xmatters.connection
import xmatters.utils

class PaginationLinks(object):
    def __init__(self, data):
        self.next = data.get('next')
        self.previous = data.get('previous')
        self.self = data.get('self')

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class Pagination(xmatters.connection.ApiBridge):

    def __init__(self, parent, data, constructor, limit=None):
        super(Pagination, self).__init__(parent, data)
        self.parent = parent
        self.constructor = constructor
        self.limit = limit
        self.state = 0  # count of items iterated
        self.total = None
        self._init_data = data

        # properties change every page
        self.count = None
        self.data = None
        self.links = None
        self.index = None
        self._set_pagination_properties(data)

    def list(self):
        return list(self)

    def goto_next_page(self):
        url = self.build_url(self.links.next)
        data = self.con.get(url)
        self._set_pagination_properties(data)

    def _set_pagination_properties(self, data):
        self.count = data.get('count')
        self.data = data.get('data')
        links = data.get('links')
        self.links = PaginationLinks(links) if links else None
        self.total = data.get('total')
        self.index = 0

    def _get_object(self, item_data):
        if issubclass(self.constructor, xmatters.utils.Factory):
            data_object = self.constructor.compose(self, item_data)
        elif issubclass(self.constructor, xmatters.connection.ApiBridge):
            data_object = self.constructor(self, item_data)
        else:
            data_object = self.constructor(item_data)
        return data_object

    def __iter__(self):
        return self

    def __next__(self):

        if (self.state == self.total) or (self.limit is not None and self.state == self.limit):
            raise StopIteration()

        if self.index == self.count and self.links and self.links.next:
            self.goto_next_page()

        if self.index < self.count:
            item_data = self.data[self.index]
            self.index += 1
            self.state += 1
            return self._get_object(item_data)

    def __len__(self):
        return self.total

    def __repr__(self):
        return '<{} {} {} objects>'.format(self.__class__.__name__, self.total, self.constructor.__name__)

    def __str__(self):
        return self.__repr__()


class Recipient(xmatters.connection.ApiBridge):
    def __init__(self, parent, data):
        super(Recipient, self).__init__(parent, data)
        self.id = data.get('id')
        self.target_name = data.get('targetName')
        self.recipient_type = data.get('recipientType')
        self.external_key = data.get('externalKey')
        self.externally_owned = data.get('externallyOwned')
        self.locked = data.get('locked')
        self.status = data.get('status')
        links = data.get('links')
        self.links = SelfLink(self, links) if links else None

    def __repr__(self):
        return '<{} {} {}>'.format(self.__class__.__name__, self.recipient_type, self.target_name)

    def __str__(self):
        return self.__repr__()


class RecipientReference(xmatters.connection.ApiBridge):
    def __init__(self, parent, data):
        super(RecipientReference, self).__init__(parent, data)
        self.id = data.get('id')
        self.target_name = data.get('targetName')
        self.recipient_type = data.get('recipientType')
        links = data.get('links')
        self.links = SelfLink(self, links) if links else None

    def __repr__(self):
        return '<{} {} {}>'.format(self.__class__.__name__, self.recipient_type, self.target_name)

    def __str__(self):
        return self.__repr__()


class SelfLink(xmatters.connection.ApiBridge):
    def __init__(self, parent, data):
        super(SelfLink, self).__init__(parent, data)
        self.self = data.get('self')

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class RecipientPointer(xmatters.connection.ApiBridge):
    def __init__(self, parent, data):
        super(RecipientPointer, self).__init__(parent, data)
        self.id = data.get('id')
        self.recipient_type = data.get('recipient')

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class ReferenceById(object):
    def __init__(self, data):
        self.id = data.get('id')

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class ReferenceByIdAndSelfLink(xmatters.connection.ApiBridge):
    def __init__(self, parent, data):
        super(ReferenceByIdAndSelfLink, self).__init__(parent, data)
        self.id = data.get('id')
        links = data.get('links')
        self.links = SelfLink(self, links) if links else None

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class PropertyDefinition(object):
    def __init__(self, data):
        self.id = data.get('id')
        self.name = data.get('name')
        self.description = data.get('description')
        self.help_text = data.get('helpText')
        self.default = data.get('default')
        self.max_length = data.get('maxLength')
        self.min_length = data.get('minLength')
        self.pattern = data.get('pattern')
        self.validate = data.get('validate')

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.name)

    def __str__(self):
        return self.__repr__()


class RequestReference(object):
    def __init__(self, data):
        self.request_id = data.get('requestId')

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()



import xmatters.xm_objects.common
import xmatters.xm_objects.plans
from xmatters.connection import ApiBridge
import xmatters.factories


class ServiceAuthentication(object):
    def __init__(self, data):
        self.username = data.get('username')
        self.connection_status = data.get('connectionStatus')

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class BasicAuthentication(object):
    def __init__(self, data):
        self.username = data.get('username')
        self.password = data.get('password')

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class OAuth2Authentication(object):
    def __init__(self, data):
        self.username = data.get('username')
        self.oauth_token_url = data.get('oauthTokenUrl')
        self.oauth_client_id = data.get('oauthClientId')
        self.client_secret = data.get('client_secret')

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()


class Endpoint(ApiBridge):
    def __init__(self, parent, data):
        super(Endpoint, self).__init__(parent, data)
        self.id = data.get('id')
        plan = data.get('plan')
        self.plan = xmatters.xm_objects.plans.PlanReference(plan) if plan else None
        self.url = data.get('url')
        self.endpoint_type = data.get('endpointType')
        self.authentication_type = data.get('authenticationType')
        auth = data.get('authentication')
        self.authentication = xmatters.factories.AuthFactory.compose(self, data) if auth else None
        links = data.get('links')
        self.links = xmatters.xm_objects.common.SelfLink(self, data) if links else None
        self.trust_self_signed = data.get('trustSelfSigned')
        self.preemptive = data.get('preemptive')
        self.data = data.get('data')

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    def __str__(self):
        return self.__repr__()
